/**************************************************************************

   Fotoxx      edit photos and manage collections

   Copyright 2007 2008 2009 2010 2011 2012 Michael Cornelison
   Source URL: http://kornelix.squarespace.com/fotoxx
   Contact: kornelix2@gmail.com
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************/

#define EX extern                                                          //  enable extern declarations
#include "fotoxx.h"


/**************************************************************************

   Fotoxx image editor - image metadata functions, EXIF/IPTC etc.

***************************************************************************/

int   get_mouse_tag(GtkTextView *, int px, int py, cchar *);               //  get tag selected by mouse
int   add_tag(char *tag, char *taglist, int maxcc);                        //  add tag if unique and enough space
int   del_tag(char *tag, char *taglist);                                   //  remove tag from tag list
int   add_recentag(char *tag);                                             //  add tag to recent tags, keep most recent
void  load_deftags();                                                      //  tags_defined file >> tags_deftags[]
void  save_deftags();                                                      //  tags_deftags[] >> defined_tags file
int   find_deftag(char *tag);                                              //  find tag in tags_deftags[]
int   add_deftag(char *catg, char *tag);                                   //  add tag to tags_deftags[]
int   del_deftag(char *tag);                                               //  remove tag from tags_deftags[]
void  deftags_stuff(zdialog *zd);                                          //  tags_deftags[] >> dialog widget "deftags"

char     meta_date[12] = "";                                               //  image date, yyyymmdd
char     meta_prdate[12] = "";                                             //  previous image date read or set
char     meta_stars = '0';                                                 //  image rating in stars, '0' to '5'
char    *tags_deftags[maxtagcats];                                         //  defined tags: catg: tag1, tag2, ... tagN,
char     tags_deftext[tagTcc] = "";                                        //  defined tags as flat text buffer
char     tags_currtags[tagFcc] = "";                                       //  tags for current image file
char     tags_recentags[tagRcc] = "";                                      //  recently added tags list
char     tags_batchAddTags[tagMcc] = "";                                   //  batch add tags list
char     tags_searchtags[tagScc] = "";                                     //  search tags list
char     tags_searchtext[tagScc] = "";                                     //  search comments & captions word list
char     meta_searchfiles[tagScc] = "";                                    //  search files list
char     meta_comments[exif_maxcc];                                        //  image comments            expanded
char     meta_caption[exif_maxcc];                                         //  image caption             v.12.06
char     meta_city[100], meta_country[100];                                //  geotags: city, country
char     meta_latitude[20], meta_longitude[20];                            //  geotags: latitude and longitude (-123.4567)

int      Ngeotags = 0;                                                     //  number of city geotag records
char     **geocity, **geocountry;                                          //  city, country, latitude, longitude
char     **geolat, **geolong;                                              //  (format +/-nnn.nnnn degrees)

zdialog  *geotags_worldmap_dialog;                                         //  active world map dialog
char     geotags_worldmap_clicklat[20];                                    //  world map clicked position,
char     geotags_worldmap_clicklong[20];                                   //    latitude, longitude (+/-nnn.nnnn)
int      Fgeotags_worldmap_data = 0;                                       //  flag, data from world map is present
int      Fgeotags_worldmap_getcity = 0;                                    //  flag, use nearest city geotags data

char     *geotags_worldmap_city;                                           //  nearest city geotag data for
char     *geotags_worldmap_country;                                        //    clicked latitude, longitude
char     *geotags_worldmap_latitude;
char     *geotags_worldmap_longitude;


/**************************************************************************/

//  View and edit IPTC Caption and EXIF UserComment data

void m_edit_cctext(GtkWidget *, cchar *menu)                               //  2 functions combined
{
   int      edit_cctext_dialog_event(zdialog *zd, cchar *event);
   char     text[exif_maxcc+50];
   cchar    *title = ZTX("Edit Caption and Comments");

   if (menu) zfuncs::F1_help_topic = "edit_cctext";

   if (! curr_file) return;

   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(mWin,Bexiftoolmissing);
      return;
   }
   
   if (! zdeditcctext)                                                     //  popup dialog if not already
   {   
      zdeditcctext = zdialog_new(title,mWin,Bapply,Bcancel,null);
      zdialog *zd = zdeditcctext;
      zdialog_add_widget(zd,"hbox","hbcap","dialog");
      zdialog_add_widget(zd,"label","labcap","hbcap","Caption","space=5");
      zdialog_add_widget(zd,"frame","framecap","hbcap",0,"expand");
      zdialog_add_widget(zd,"edit","textcap","framecap",0,"space=5|wrap");    
      zdialog_add_widget(zd,"hbox","hbcom","dialog");
      zdialog_add_widget(zd,"label","labcom","hbcom","Comments","space=5");
      zdialog_add_widget(zd,"frame","framecom","hbcom",0,"expand");
      zdialog_add_widget(zd,"edit","textcom","framecom",0,"space=5|wrap");    
      zdialog_resize(zd,600,0);
      zdialog_help(zd,"edit_cctext");                                      //  zdialog help topic
      zdialog_run(zd,edit_cctext_dialog_event);
   }
   
   load_filemeta(curr_file);                                               //  get current caption and comments
   repl_1str(meta_caption,text,"\\n","\n");                                //  replace "\n" with real newlines
   zdialog_stuff(zdeditcctext,"textcap",text);                             //  stuff into dialog
   repl_1str(meta_comments,text,"\\n","\n");                               //  replace "\n" with real newlines
   zdialog_stuff(zdeditcctext,"textcom",text);                             //  stuff into dialog

   return;
}


//  dialog event and completion callback function

int  edit_cctext_dialog_event(zdialog *zd, cchar *event)
{
   char     text[exif_maxcc+50];
   
   if (! zd->zstat) return 0;
   
   else if (zd->zstat == 1)                                                //  apply, save text
   {
      zd->zstat = 0;                                                       //  keep dialog active
      zdialog_fetch(zd,"textcap",text,exif_maxcc);                         //  get new cctext
      repl_1str(text,meta_caption,"\n","\\n");                             //  replace newlines with "\n"
      zdialog_fetch(zd,"textcom",text,exif_maxcc);                         //  get new cctext
      repl_1str(text,meta_comments,"\n","\\n");                            //  replace newlines with "\n"
      Ftagschanged = 1;
      save_filemeta(curr_file);                                            //  save in image file
   }

   else zdialog_free(zdeditcctext);                                        //  cancel

   return 1;
}


/**************************************************************************/

//  edit tags menu function

void m_edit_tags(GtkWidget *, cchar *menu)
{
   void  edit_filetags_clickfunc(GtkWidget *widget, int &line, int &pos);
   void  edit_recentags_clickfunc(GtkWidget *widget, int &line, int &pos);
   void  edit_deftags_clickfunc(GtkWidget *widget, int &line, int &pos);
   int   edittags_dialog_event(zdialog *zd, cchar *event);

   char        *ppv, starsN[12];
   cchar       *mantags = ZTX("Manage Tags");
   GtkWidget   *widget;
   zdialog     *zd;

   if (menu) zfuncs::F1_help_topic = "edit_tags";

   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(mWin,Bexiftoolmissing);
      return;
   }
   
   if (! curr_file) return;

   if (! zdedittags)                                                       //  (re) start tag edit dialog 
   {
      zd = zdialog_new(ZTX("Edit Tags"),mWin,mantags,Bapply,Bcancel,null);
      zdedittags = zd;

      zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
      zdialog_add_widget(zd,"label","labfile","hb1",ZTX("file:"),"space=5");
      zdialog_add_widget(zd,"label","file","hb1");

      zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
      zdialog_add_widget(zd,"label","lab21","hb2",ZTX("image date (yyyymmdd)"),"space=5");
      zdialog_add_widget(zd,"entry","date","hb2",0,"scc=12");
      zdialog_add_widget(zd,"button","prdate","hb2",ZTX("use last"),"space=5");

      zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
      zdialog_add_widget(zd,"label","labstars","hb3",ZTX("image stars"),"space=5");
      zdialog_add_widget(zd,"vbox","vb3","hb3");
      zdialog_add_widget(zd,"hbox","hb31","vb3",0,"homog");
      zdialog_add_widget(zd,"hbox","hb32","vb3",0,"homog");
      zdialog_add_widget(zd,"label","lab30","hb31","0");
      zdialog_add_widget(zd,"label","lab31","hb31","1");
      zdialog_add_widget(zd,"label","lab32","hb31","2");
      zdialog_add_widget(zd,"label","lab33","hb31","3");
      zdialog_add_widget(zd,"label","lab34","hb31","4");
      zdialog_add_widget(zd,"label","lab35","hb31","5");
      zdialog_add_widget(zd,"radio","stars0","hb32",0);
      zdialog_add_widget(zd,"radio","stars1","hb32",0);
      zdialog_add_widget(zd,"radio","stars2","hb32",0);
      zdialog_add_widget(zd,"radio","stars3","hb32",0);
      zdialog_add_widget(zd,"radio","stars4","hb32",0);
      zdialog_add_widget(zd,"radio","stars5","hb32",0);

      zdialog_add_widget(zd,"hbox","hb4","dialog",0,"space=5");
      zdialog_add_widget(zd,"label","lab4","hb4",ZTX("current tags"),"space=5");
      zdialog_add_widget(zd,"frame","frame4","hb4",0,"expand");
      zdialog_add_widget(zd,"edit","filetags","frame4",0,"expand|wrap");

      zdialog_add_widget(zd,"hbox","hb5","dialog",0,"space=5");
      zdialog_add_widget(zd,"label","recent","hb5",ZTX("recent tags"),"space=5");
      zdialog_add_widget(zd,"frame","frame5","hb5",0,"expand");
      zdialog_add_widget(zd,"edit","recentags","frame5",0,"expand|wrap");

      zdialog_add_widget(zd,"hbox","hb8","dialog");
      zdialog_add_widget(zd,"label","labdeftags","hb8",ZTX("defined tags"),"space=5");
      zdialog_add_widget(zd,"hbox","hb9","dialog",0,"expand");
      zdialog_add_widget(zd,"frame","frame8","hb9",0,"space=5|expand");
      zdialog_add_widget(zd,"scrwin","scrwin8","frame8",0,"expand");
      zdialog_add_widget(zd,"edit","deftags","scrwin8",0,"expand|wrap");

      load_deftags();                                                      //  stuff defined tags into dialog
      deftags_stuff(zd);

      widget = zdialog_widget(zd,"filetags");                              //  tag widget mouse functions   v.12.06
      textwidget_set_clickfunc(widget,edit_filetags_clickfunc);

      widget = zdialog_widget(zd,"recentags");
      textwidget_set_clickfunc(widget,edit_recentags_clickfunc);

      widget = zdialog_widget(zd,"deftags");
      textwidget_set_clickfunc(widget,edit_deftags_clickfunc);
      
      zdialog_resize(zd,500,500);                                          //  run dialog
      zdialog_help(zd,"edit_tags");                                        //  zdialog help topic 
      zdialog_run(zd,edittags_dialog_event);
   }
   
   zd = zdedittags;

   load_filemeta(curr_file);                                               //  get EXIF/IPTC data

   ppv = (char *) strrchr(curr_file,'/');
   zdialog_stuff(zd,"file",ppv+1);                                         //  stuff dialog file name

   zdialog_stuff(zd,"date",meta_date);                                     //  stuff dialog data
   sprintf(starsN,"stars%c",meta_stars);
   zdialog_stuff(zd,starsN,1);
   zdialog_stuff(zd,"filetags",tags_currtags);
   zdialog_stuff(zd,"recentags",tags_recentags);

   return;
}


//  mouse click functions for various text widgets for tags                //  v.12.06

void edit_filetags_clickfunc(GtkWidget *widget, int &line, int &pos)       //  existing image tag was clicked
{
   char     *txline, *txtag, end = 0;
   cchar    *delims = tagdelims":";
   
   txline = textwidget_get_line(widget,line);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,delims,end);
   if (! txtag) { zfree(txline); return; }
   
   del_tag(txtag,tags_currtags);                                           //  remove tag from image
   zdialog_stuff(zdedittags,"filetags",tags_currtags);
   
   zfree(txline);
   zfree(txtag);
   return;
}


void edit_recentags_clickfunc(GtkWidget *widget, int &line, int &pos)      //  recent tag was clicked
{
   char     *txline, *txtag, end = 0;
   cchar    *delims = tagdelims":";
   
   txline = textwidget_get_line(widget,line);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,delims,end);
   if (! txtag) { zfree(txline); return; }
   
   add_tag(txtag,tags_currtags,tagFcc);                                    //  add recent tag to image
   zdialog_stuff(zdedittags,"filetags",tags_currtags);
   
   zfree(txline);
   zfree(txtag);
   return;
}


void edit_deftags_clickfunc(GtkWidget *widget, int &line, int &pos)        //  defined tag was clicked
{
   char     *txline, *txtag, end = 0;
   cchar    *delims = tagdelims":";
   
   txline = textwidget_get_line(widget,line);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,delims,end);
   if (! txtag || end == ':') { zfree(txline); return; }                   //  tag category clicked, ignore
   
   add_tag(txtag,tags_currtags,tagFcc);                                    //  add new tag to image
   zdialog_stuff(zdedittags,"filetags",tags_currtags);                     //    from defined tags list

   add_recentag(txtag);                                                    //  and add to recent tags
   zdialog_stuff(zdedittags,"recentags",tags_recentags);
   
   zfree(txline);
   zfree(txtag);
   return;
}


//  dialog event and completion callback function

int edittags_dialog_event(zdialog *zd, cchar *event)
{
   int         err;
   
   if (zd->zstat) {
      if (zd->zstat == 1) {                                                //  manage tags
         zd->zstat = 0;                                                    //  keep dialog active
         manage_tags();
         return 0;
      }

      if (zd->zstat == 2) {                                                //  [apply]
         zd->zstat = 0;                                                    //  keep dialog active
         save_filemeta(curr_file);                                         //  save tag changes
         return 0;
      }

      zdialog_free(zdedittags);                                            //  cancel - kill dialog
      return 0;
   }
   
   if (strEqu(event,"date")) {                                             //  image date revised
      err = zdialog_fetch(zd,"date",meta_date,11);
      if (err) return 1;
      if (strlen(meta_date) == 4) strcat(meta_date,"0101");                //  yyyy >> yyyy0101
      if (strlen(meta_date) == 6) strcat(meta_date,"01");                  //  yyyymm >> yyyymm01
      Ftagschanged++;
   }
   
   if (strEqu(event,"prdate")) {                                           //  repeat last date used
      if (*meta_prdate) {
         zdialog_stuff(zd,"date",meta_prdate);
         strcpy(meta_date,meta_prdate);
         Ftagschanged++;
      }
   }

   if (strnEqu(event,"stars",5)) {                                         //  event = stars0 to stars5
      meta_stars = event[5];                                               //  '0' to '5'
      Ftagschanged++;
   }
   
   if (strEqu(event,"tags-changed"))                                       //  get new defined tags data
      deftags_stuff(zdedittags);

   return 0;
}


/**************************************************************************/

//  manage tags function - dialog

zdialog  *zdmanagetags = 0;

void manage_tags()
{
   void  manage_deftags_clickfunc(GtkWidget *widget, int &line, int &pos);
   int   managetags_dialog_event(zdialog *zd, cchar *event);

   GtkWidget   *widget;
   zdialog     *zd;

   if (zdmanagetags) return;
   zd = zdialog_new(ZTX("Manage Tags"),mWin,Bcancel,null);
   zdmanagetags = zd;

   zdialog_add_widget(zd,"hbox","hb7","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labcatg","hb7",ZTX("category"),"space=5");
   zdialog_add_widget(zd,"entry","catg","hb7",0,"scc=12");
   zdialog_add_widget(zd,"label","space","hb7",0,"space=5");
   zdialog_add_widget(zd,"label","labtag","hb7",ZTX("tag"),"space=5");
   zdialog_add_widget(zd,"entry","tag","hb7",0,"scc=20|expand");
   zdialog_add_widget(zd,"label","space","hb7",0,"space=5");
   zdialog_add_widget(zd,"button","create","hb7",ZTX("create"));
   zdialog_add_widget(zd,"button","delete","hb7",ZTX("delete"));

   zdialog_add_widget(zd,"hbox","hb8","dialog");
   zdialog_add_widget(zd,"label","labdeftags","hb8",ZTX("defined tags"),"space=5");
   zdialog_add_widget(zd,"hbox","hb9","dialog",0,"expand");
   zdialog_add_widget(zd,"frame","frame8","hb9",0,"space=5|expand");
   zdialog_add_widget(zd,"scrwin","scrwin8","frame8",0,"expand");
   zdialog_add_widget(zd,"edit","deftags","scrwin8",0,"expand|wrap");

   widget = zdialog_widget(zd,"deftags");                                  //  tag widget mouse function    v.12.06
   textwidget_set_clickfunc(widget,manage_deftags_clickfunc);

   load_deftags();                                                         //  stuff defined tags into dialog
   deftags_stuff(zd);

   zdialog_resize(zd,0,400);
   zdialog_help(zd,"edit_tags");                                           //  zdialog help topic 
   zdialog_run(zd,managetags_dialog_event);                                //  run dialog, parallel
   
   return;
}


//  mouse click functions for widget having tags                           //  v.12.06

void manage_deftags_clickfunc(GtkWidget *widget, int &line, int &pos)      //  tag or tag category was clicked
{
   char     *txline, *txtag, end = 0;
   cchar    *delims = tagdelims":";
   
   txline = textwidget_get_line(widget,line);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,delims,end);
   if (! txtag) { zfree(txline); return; }
   
   if (end == ':')                                                         //  tag category >> dialog widget
      zdialog_stuff(zdmanagetags,"catg",txtag);
   else 
      zdialog_stuff(zdmanagetags,"tag",txtag);                             //  selected tag >> dialog widget
   
   zfree(txline);
   zfree(txtag);
   return;
}


//  dialog event and completion callback function

int managetags_dialog_event(zdialog *zd, cchar *event)
{
   char        tag[tagcc], catg[tagcc];
   int         changed = 0;
   
   if (zd->zstat) {
      zdialog_free(zd);
      return 0;
   }
   
   if (strEqu(event,"create")) {                                           //  add new tag to defined tags
      zdialog_fetch(zd,"catg",catg,tagcc);
      zdialog_fetch(zd,"tag",tag,tagcc);
      add_deftag(catg,tag);
      changed++;
   }

   if (strEqu(event,"delete")) {                                           //  remove tag from defined tags
      zdialog_fetch(zd,"tag",tag,tagcc);
      del_deftag(tag);
      changed++;
   }

   if (changed) {   
      save_deftags();                                                      //  save tag updates to file
      deftags_stuff(zd);                                                   //  update dialog "deftags" window
      if (zdedittags)                                                      //  and edit tags window if active
         zdialog_send_event(zdedittags,"tags-changed");
   }

   return 0;
}


/**************************************************************************/

//  add input tag to output tag list if not already there and enough room
//  returns:   0 = added OK     1 = not unique (case ignored)
//             2 = overflow     3 = bad utf8 characters     4 = null tag

int add_tag(char *tag, char *taglist, int maxcc)
{
   char     *pp1, *pp2, tag1[tagcc], tag2[tagcc];
   int      cc, cc1, cc2;

   strncpy0(tag1,tag,tagcc);                                               //  remove leading and trailing blanks
   cc = strTrim2(tag2,tag1);
   if (! cc) return 4;

   if (utf8_check(tag2)) {                                                 //  check for valid utf8 encoding
      printf("bad utf8 characters: %s \n",tag2);
      return 3;
   }
   
   while ((pp1 = strpbrk(tag2,tagdelims":"))) *pp1 = '-';                  //  replace problem characters

   strcpy(tag,tag2);                                                       //  replace tag with sanitized version
   
   pp1 = taglist;
   cc1 = strlen(tag);

   while (true)                                                            //  check if already in tag list
   {
      while (*pp1 == ' ' || *pp1 == tagdelim1) pp1++;
      if (! *pp1) break;
      pp2 = pp1 + 1;
      while (*pp2 && *pp2 != tagdelim1) pp2++;
      cc2 = pp2 - pp1;
      if (cc2 == cc1 && strncaseEqu(tag,pp1,cc1)) return 1;
      pp1 = pp2;
   }

   cc2 = strlen(taglist);                                                  //  append to tag list if space enough
   if (cc1 + cc2 + 3 > maxcc) return 2;
   strcpy(taglist + cc2,tag);
   strcpy(taglist + cc2 + cc1, tagdelim2);                                 //  add delimiter + space

   if (taglist == tags_currtags)                                           //  image tags were changed
      Ftagschanged++;
   return 0;
}


//  remove tag from taglist, if present
//  returns: 0 if found and deleted, otherwise 1

int del_tag(char *tag, char *taglist)
{
   int         ii, ftcc, atcc, found;
   char        *temptags;
   cchar       *pp;
   
   temptags = strdupz(taglist,0,"del_tag");
   
   *taglist = 0;
   ftcc = found = 0;
   
   for (ii = 1; ; ii++)
   {
      pp = strField(temptags,tagdelims,ii);                                //  next tag
      if (! pp) {
         zfree(temptags);
         if (found && taglist == tags_currtags)                            //  image tags were changed
            Ftagschanged++;
         return 1-found;
      }
      if (*pp == ' ') continue;
      
      if (strcaseEqu(pp,tag)) {                                            //  skip matching tag
         found = 1;
         continue;
      }

      atcc = strlen(pp);                                                   //  copy non-matching tag
      strcpy(taglist + ftcc, pp);
      ftcc += atcc;
      strcpy(taglist + ftcc, tagdelim2);                                   //  + delim + blank
      ftcc += 2;
   }
}


//  add new tag to recent tags, if not already.
//  remove oldest to make space if needed.

int add_recentag(char *tag)
{
   int         err;
   char        *pp, temptags[tagRcc];

   err = add_tag(tag,tags_recentags,tagRcc);                               //  add tag to recent tags

   while (err == 2)                                                        //  overflow
   {
      strncpy0(temptags,tags_recentags,tagRcc);                            //  remove oldest to make room
      pp = strpbrk(temptags,tagdelims);
      if (! pp) return 0;
      strcpy(tags_recentags,pp+2);                                         //  delimiter + blank before tag
      err = add_tag(tag,tags_recentags,tagRcc);
   }

   return 0;
}


/**************************************************************************/

//  Load tags_defined file into tags_deftags[ii] => category: tag1, tag2, ...
//  Read search_index file and add unmatched tags: => nocatg: tag1, tag2, ...

void load_deftags()
{
   int tags_Ucomp(cchar *tag1, cchar *tag2);

   static int  Floaded = 0;
   FILE        *fid;
   int         ii, jj, ntags, err, cc, tcc;
   int         ncats, catoverflow;
   int         nocat, nocatcc;
   char        tag[tagcc], catg[tagcc];
   char        tagsbuff[tagGcc];
   char        *pp1, *pp2;
   char        ptags[tagntc][tagcc];
   
   if (Floaded) return;                                                    //  use memory tags if already there
   Floaded++;

   for (ii = 0; ii < maxtagcats; ii++)                                     //  clean memory
      tags_deftags[ii] = 0;

   ncats = catoverflow = 0;

   fid = fopen(tags_defined_file,"r");                                     //  read tags_defined file
   if (fid) {
      while (true) {
         pp1 = fgets_trim(tagsbuff,tagGcc,fid);
         if (! pp1) break;
         if (ncats == maxtagcats-1) goto toomanycats;
         pp2 = strchr(pp1,':');                                            //  isolate "category:"
         if (! pp2) continue;                                              //  reject bad data
         cc = pp2 - pp1 + 1;
         if (cc > tagcc-1) continue;
         strncpy0(catg,pp1,cc);                                            //  (for error message)
         if (strlen(pp1) > tagGcc-2) goto cattoobig;
         pp2++;
         while (*pp2 == ' ') pp2++;
         if (strlen(pp2) < 3) continue;
         while ((pp2 = strpbrk(pp2,tagdelims))) *pp2++ = tagdelim1;        //  force comma delimiter
         tags_deftags[ncats] = strdupz(pp1,0,"def_tags");                  //  tags_deftags[ii]
         ncats++;                                                          //   = category: tag1, tag2, ... tagN,
      }
      err = fclose(fid);
      if (err) goto deftagserr;
   }

   nocat = ncats;                                                          //  make last category "nocatg" for
   ncats++;                                                                //   unmatched tags in search_index file
   tags_deftags[nocat] = zmalloc(tagGcc,"deftags");
   strcpy(tags_deftags[nocat],"nocatg: ");
   nocatcc = 8;

   fid = fopen(search_index_file,"r");                                     //  read search_index file
   if (! fid) return;

   while (true)
   {
      pp1 = fgets_trim(tagsbuff,tagGcc,fid);                               //  next image file tags record
      if (! pp1) break;
      if (strnNeq(tagsbuff,"tags: ",6)) continue;
      pp1 = pp1 + 6;
      
      while (true)
      {
         while (*pp1 && strchr(tagdelims" ",*pp1)) pp1++;                  //  next image tag start
         if (! *pp1) break;
         pp2 = strpbrk(pp1,tagdelims);                                     //  end
         if (! pp2) pp2 = pp1 + strlen(pp1);  
         cc = pp2 - pp1;
         if (cc > tagcc-1) {
            pp1 = pp2;
            continue;                                                      //  ignore huge tag
         }
         
         strncpy0(tag,pp1,cc+1);                                           //  look for tag in defined tags
         err = find_deftag(tag);
         if (! err) {                                                      //  found
            pp1 = pp2;
            continue;
         }

         if (nocatcc + cc + 2 > tagGcc-2) {
            catoverflow = 1;                                               //  nocatg: length limit reached
            break;
         }
         else {
            strcpy(tags_deftags[nocat] + nocatcc, tag);                    //  append tag to list
            nocatcc += cc;
            strcpy(tags_deftags[nocat] + nocatcc, tagdelim2);              //  + delim + blank
            nocatcc += 2;
         }

         pp1 = pp2;
      }
   }

   err = fclose(fid);
   if (err) goto filetagserr;
   if (catoverflow) goto cattoobig;
   
//  parse all the tags in each category and sort in ascending order

   for (ii = 0; ii < ncats; ii++)
   {
      pp1 = tags_deftags[ii];
      pp2 = strchr(pp1,':');
      cc = pp2 - pp1 + 1;
      strncpy0(catg,pp1,cc);
      pp1 = pp2 + 1;
      while (*pp1 == ' ') pp1++;
      tcc = 0;

      for (jj = 0; jj < tagntc; jj++)
      {
         if (! *pp1) break;
         pp2 = strchr(pp1,tagdelim1);
         if (pp2) cc = pp2 - pp1;
         else cc = strlen(pp1);
         if (cc > tagcc-1) cc = tagcc-1;
         strncpy0(ptags[jj],pp1,cc+1);
         pp1 += cc + 1;
         tcc += cc;
         while (*pp1 == ' ') pp1++;
      }
      
      ntags = jj;
      if (ntags == tagntc) goto cattoobig;
      HeapSort((char *) ptags,tagcc,ntags,tags_Ucomp);

      pp1 = tags_deftags[ii];
      tcc += strlen(catg) + 2 + 2 * ntags + 2;                             //  category, all tags, delimiters
      pp2 = zmalloc(tcc,"deftags");

      tags_deftags[ii] = pp2;                                              //  swap memory
      zfree(pp1);

      strcpy(pp2,catg);
      pp2 += strlen(catg);
      strcpy(pp2,": ");                                                    //  pp2 = "category: "
      pp2 += 2;

      for (jj = 0; jj < ntags; jj++)                                       //  add the sorted tags
      {
         strcpy(pp2,ptags[jj]);                                            //  append tag + delim + blank
         pp2 += strlen(pp2);
         strcpy(pp2,tagdelim2);
         pp2 += 2;
      }
      
      *pp2 = 0;
   }
   
//  sort the categories in ascending order
//  leave "nocatg" at the end

   for (ii = 0; ii < ncats-1; ii++)
   for (jj = ii+1; jj < ncats-1; jj++) 
   {
      pp1 = tags_deftags[ii];
      pp2 = tags_deftags[jj];
      if (strcasecmp(pp1,pp2) > 0) {
         tags_deftags[ii] = pp2;
         tags_deftags[jj] = pp1;
      }
   }

   return;

toomanycats:
   zmessLogACK(mWin,"more than %d categories",maxtagcats);
   fclose(fid);
   return;
   
cattoobig:
   zmessLogACK(mWin,"category %s is too big",catg);
   fclose(fid);
   return;

deftagserr:
   zmessLogACK(mWin,"tags_defined file error: %s",strerror(errno));
   return;

filetagserr:
   zmessLogACK(mWin,"search_index file error: %s",strerror(errno));
   return;
}


//  compare function for tag sorting

int tags_Ucomp(cchar *tag1, cchar *tag2)
{
   return strcasecmp(tag1,tag2);
}


//  write tags_deftags[] memory data to the defined tags file if any changes were made

void save_deftags()
{
   int         ii, err;
   FILE        *fid;

   fid = fopen(tags_defined_file,"w");                                     //  write tags_defined file
   if (! fid) goto deftagserr;

   for (ii = 0; ii < maxtagcats; ii++)
   {
      if (! tags_deftags[ii+1]) break;                                     //  omit last category, "nocatg"
      err = fprintf(fid,"%s\n",tags_deftags[ii]);                          //  each record: 
      if (err < 0) goto deftagserr;                                        //    category: tag1, tag2, ... tagN,
   }

   err = fclose(fid);
   if (err) goto deftagserr;
   return;
   
deftagserr:
   zmessLogACK(mWin,"tags_defined file error: %s",strerror(errno));
   return;
}


//  find a given tag in tags_deftags[]
//  return: 0 = found, 1 = not found

int find_deftag(char *tag)
{
   int      ii, cc;
   char     tag2[tagcc+4];
   char     *pp;

   if (! tag || *tag <= ' ') return 0;                                     //  bad tag

   strncpy0(tag2,tag,tagcc);                                               //  construct tag + delim + blank
   cc = strlen(tag2);
   strcpy(tag2+cc,tagdelim2);
   cc += 2;

   for (ii = 0; ii < maxtagcats; ii++)
   {
      pp = tags_deftags[ii];                                               //  category: tag1, tag2, ... tagN, 
      if (! pp) return 1;                                                  //  not found

      while (pp)
      {
         pp = strcasestr(pp,tag2);                                         //  look for delim + blank + tag + delim
         if (! pp) break;
         if (strchr(tagdelims":", pp[-2])) return 0;
         pp += cc;
      }
   }

   return 1;
}


//  add new tag to tags_deftags[] >> category: tag1, tag2, ... newtag,
//  returns:   0 = added OK     1 = not unique (case ignored)
//             2 = overflow     3 = bad utf8 characters     4 = null tag
//  if tag present under another category, it is moved to new category

int add_deftag(char *catg, char *tag)
{
   int         ii, cc, cc1, cc2;
   char        tag1[tagcc], tag2[tagcc];
   char        *pp1, *pp2;
   
   strncpy0(tag1,tag,tagcc);                                               //  remove leading and trailing blanks
   cc = strTrim2(tag2,tag1);
   if (! cc) return 4;

   if (utf8_check(tag2)) {                                                 //  check for valid utf8 encoding
      printf("bad utf8 characters: %s \n",tag2);
      return 3;
   }

   while ((pp1 = strpbrk(tag2,tagdelims":"))) *pp1 = '-';                  //  replace problem characters
   
   strcpy(tag,tag2);                                                       //  replace tag with sanitized version

   del_deftag(tag);                                                        //  delete if already there

   if (! catg || *catg <= ' ') catg = (char *) "nocatg";

   cc1 = strlen(catg);

   for (ii = 0; ii < maxtagcats; ii++)                                     //  look for given category
   {
      pp1 = tags_deftags[ii];
      if (! pp1) goto newcatg;
      if (! strnEqu(catg,pp1,cc1)) continue;
      if (pp1[cc1] == ':') goto oldcatg;
   }

newcatg:
   if (ii == maxtagcats) goto toomanycats;
   cc1 = strlen(catg) + strlen(tag) + 6;
   pp1 = zmalloc(cc1,"deftags");
   *pp1 = 0;
   strncatv(pp1,cc1,catg,": ",tag,tagdelim2,null);                         //  category: + tag + delim + blank
   tags_deftags[ii] = tags_deftags[ii-1];                                  //  move "nocatg" record to next slot
   tags_deftags[ii-1] = pp1;                                               //  insert new record before
   return 0;

oldcatg:
   cc1 = strlen(tag);
   pp2 = pp1 + 2;
   while (true) {
      pp2 = strcasestr(pp2,tag);                                           //  look for delim + blank + tag + delim
      if (! pp2) break;                                                    //        or colon + blank + tag + delim
      if (strchr(tagdelims,pp2[cc]) && strchr(tagdelims":", pp2[-2]))
         return 1;                                                         //  tag not unique
      pp2 += cc1;
   }
   cc2 = strlen(pp1);                                                      //  add new tag to old record
   if (cc1 + cc2 + 4 > tagGcc) goto cattoobig;
   pp2 = strdupz(pp1,cc2+4,"def_tags");                                    //  expand string
   zfree(pp1);
   tags_deftags[ii] = pp2;
   strcpy(pp2+cc2,tag);                                                    //  old record + tag + delim + blank
   strcpy(pp2+cc2+cc1,tagdelim2);
   return 0;

toomanycats:
   zmessLogACK(mWin,"more than %d categories",maxtagcats);
   return 2;

cattoobig:
   zmessLogACK(mWin,"category is too big");
   return 2;
}


//  delete tag from defined tags list, tags_deftags[] 
//  return: 0 = found and deleted, 1 = not found

int del_deftag(char *tag)
{
   int      ii, cc;
   char     tag2[tagcc+4];
   char     *pp, *pp1, *pp2;

   if (! tag || *tag <= ' ') return 1;                                     //  bad tag

   strncpy0(tag2,tag,tagcc);                                               //  construct tag + delim + blank
   cc = strlen(tag2);
   strcpy(tag2+cc,tagdelim2);
   cc += 2;

   for (ii = 0; ii < maxtagcats; ii++)
   {
      pp = tags_deftags[ii];
      if (! pp) return 1;                                                  //  not found

      while (pp)
      {
         pp = strcasestr(pp,tag2);                                         //  look for prior delim or colon
         if (! pp) break;
         if (strchr(tagdelims":", pp[-2])) goto found;
         pp += cc;
      }
   }
   
found:
   for (pp1 = pp, pp2 = pp+cc; *pp2; pp1++, pp2++)                         //  eliminate tag, delim, blank
      *pp1 = *pp2;
   *pp1 = 0;
   
   return 0;
}


//  stuff tags_deftags[] into given text widget and format by category
//  create tags_deftext with flat list of tags for mouse clicking

void deftags_stuff(zdialog *zd)
{
   GtkWidget      *widget;
   GtkTextBuffer  *textbuff;
   GtkTextIter    iter1, iter2;
   int            ii, cc;
   char           catgname[tagcc+3];
   char           *pp1, *pp2;
   
   widget = zdialog_widget(zd,"deftags");
   wclear(widget);

   for (ii = 0; ii < maxtagcats; ii++)
   {
      pp1 = tags_deftags[ii];
      if (! pp1) break;
      pp2 = strchr(pp1,':');
      if (! pp2) continue;
      if (pp2 > pp1 + tagcc-3) continue;
      pp2 += 2;
      cc = pp2 - pp1 + 1;
      strncpy0(catgname,pp1,cc);
      wprintx(widget,0,catgname,"monospace bold 9");
      if (*pp2) wprintx(widget,0,pp2,"monospace 9");
      wprintx(widget,0,"\n");
   }
   
   textbuff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget));
   gtk_text_buffer_get_bounds(textbuff,&iter1,&iter2);
   pp1 = gtk_text_buffer_get_text(textbuff,&iter1,&iter2,0);
   strncpy0(tags_deftext,pp1,tagTcc);

   return;
}


/**************************************************************************/

//  image file EXIF/IPTC data >> memory data:
//    meta_date, meta_stars, tags_currtags, meta_comments, meta_caption,
//    meta_city, meta_country, meta_latitude, meta_longitude

void load_filemeta(cchar *file)
{
   void exif_tagdate(cchar *exifdate, char *tagdate);

   int      ii, jj, cc;
   char     *pp;
   cchar    *exifkeys[9] = { exif_date_key, iptc_keywords_key, iptc_rating_key, 
                             exif_comment_key, iptc_caption_key,
                             exif_city_key, exif_country_key, 
                             exif_latitude_key, exif_longitude_key };

   char     **ppv, *imagedate, *imagekeywords, *imagestars, *imagecomms, *imagecapt;
   char     *imagecity, *imagecountry, *imagelatitude, *imagelongitude;

   *tags_currtags = *meta_date = *meta_comments = *meta_caption = 0;
   meta_stars = '0';
   *meta_city = *meta_country = *meta_latitude = *meta_longitude = 0;
   
   ppv = exif_get(file,exifkeys,9);                                        //  get metadata from image file
   imagedate = ppv[0];
   imagekeywords = ppv[1];
   imagestars = ppv[2];
   imagecomms = ppv[3];
   imagecapt = ppv[4];
   imagecity = ppv[5];
   imagecountry = ppv[6];
   imagelatitude = ppv[7];
   imagelongitude = ppv[8];
   
   if (imagedate) {
      exif_tagdate(imagedate,meta_date);                                   //  EXIF date/time >> yyyymmdd
      strcpy(meta_prdate,meta_date);
      zfree(imagedate);
   }

   if (imagekeywords)
   {
      for (ii = 1; ; ii++)
      {
         pp = (char *) strField(imagekeywords,tagdelims,ii);
         if (! pp) break;
         if (*pp == ' ') continue;
         cc = strlen(pp);
         if (cc >= tagcc) continue;                                        //  reject tags too big
         for (jj = 0; jj < cc; jj++)
            if (pp[jj] > 0 && pp[jj] < ' ') break;                         //  reject tags with control characters
         if (jj < cc) continue;
         add_tag(pp,tags_currtags,tagFcc);                                 //  add to file tags if unique
      }

      zfree(imagekeywords);
   }

   if (imagestars) {
      meta_stars = *imagestars;
      if (meta_stars < '0' || meta_stars > '5') meta_stars = '0';
      zfree(imagestars);
   }
   
   if (imagecomms) {
      strncpy0(meta_comments,imagecomms,exif_maxcc);
      zfree(imagecomms);
   }

   if (imagecapt) {
      strncpy0(meta_caption,imagecapt,exif_maxcc);
      zfree(imagecapt);
   }
   
   if (imagecity) {                                                        //  geotags    v.12.05
      strncpy0(meta_city,imagecity,99);
      zfree(imagecity);
   }

   if (imagecountry) {
      strncpy0(meta_country,imagecountry,99);
      zfree(imagecountry);
   }

   if (imagelatitude) {
      strncpy0(meta_latitude,imagelatitude,12);
      zfree(imagelatitude);
   }

   if (imagelongitude) {
      strncpy0(meta_longitude,imagelongitude,12);
      zfree(imagelongitude);
   }
   
   Ftagschanged = 0;
   return;
}


//  add metadata in memory to file EXIF/IPTC data and search_index file
//  update defined tags file if any changes

void save_filemeta(cchar *file)
{
   void tag_exifdate(cchar *tagdate, char *exifdate);
   
   cchar    *exifkeys[9] = { exif_date_key, iptc_keywords_key, iptc_rating_key, 
                             exif_comment_key, iptc_caption_key,
                             exif_city_key, exif_country_key, 
                             exif_latitude_key, exif_longitude_key };
   cchar       *exifdata[9];
   char        imagedate[24], sstars[4];

   if (! Ftagschanged) return;                                             //  no changes to tags etc.
   Ftagschanged = 0;

   *imagedate = 0;
   if (*meta_date) {
      tag_exifdate(meta_date,imagedate);                                   //  yyyymmdd >> EXIF date/time
      strcpy(meta_prdate,meta_date);
   }
   
   sstars[0] = meta_stars;                                                 //  string for stars rating
   sstars[1] = 0;

   exifdata[0] = imagedate;                                                //  update file EXIF/IPTC data
   exifdata[1] = tags_currtags;
   exifdata[2] = sstars;
   exifdata[3] = meta_comments;
   exifdata[4] = meta_caption;
   exifdata[5] = meta_city;                                                //  geotags       v.12.05
   exifdata[6] = meta_country;
   exifdata[7] = meta_latitude;
   exifdata[8] = meta_longitude;
   
   exif_put(file,exifkeys,exifdata,9);

   update_search_index(file);                                              //  update search index file

   if (zdexifview) meta_view(0);                                           //  live EXIF/IPTC update
   return;
}


//  update search index file (replace updated file data)                   //  overhauled    v.12.01

void update_search_index(cchar *file)
{
   char           *ppv, temp_search_index_file[200];
   char           imagedate[12], filedate[16];
   char           buff[indexrecl];
   int            err, eof = 0, fcopy = 0, comp = 0, finserted = 0;
   FILE           *fidr, *fidw;
   struct stat    statb;
   struct tm      bdt;

   strcpy(temp_search_index_file,search_index_file);                       //  temp tag file
   strcat(temp_search_index_file,"_temp");

   fidw = fopen(temp_search_index_file,"w");                               //  write temp tag file (new)
   if (! fidw) goto tagserror;
   
   fidr = fopen(search_index_file,"r");                                    //  read tag file (old)
   
   if (fidr)
   {   
      while (true)                                                         //  copy search index file to temp
      {                                                                    //    file, inserting or replacing
         ppv = fgets_trim(buff,indexrecl,fidr);                            //      records for caller's file
         if (! ppv) eof = 1;

         if (strnEqu(buff,"file: ",6))                                     //  start of next file entry
         {
            comp = strcmp(file,buff+6);
            if (comp > 0) fcopy = 1;                                       //  new entry is after this one
            if (comp == 0) fcopy = 0;                                      //  new entry replaces this one
            if (comp < 0) fcopy = 1;                                       //  new entry is before this one
         }

         if (comp <= 0 && ! finserted)                                     //  new entry inserts here    v.12.01
         {
            err = fprintf(fidw,"file: %s\n",file);                         //  output filespec
            if (err <= 0) goto tagserror;

            if (*meta_date) {                                              //  image date
               strncpy(imagedate,meta_date,4);
               strncpy(imagedate+5,meta_date+4,2);                         //  yyyymmdd >> yyyy:mm:dd
               strncpy(imagedate+8,meta_date+6,2);
               imagedate[4] = imagedate[7] = ':';
               imagedate[10] = 0;
            }
            else strcpy(imagedate,"null");                                 //  missing image date

            err = stat(file,&statb);
            gmtime_r(&statb.st_mtime,&bdt);
            sprintf(filedate,"%04d%02d%02d%02d%02d%02d",                   //  file date = yyyymmddhhmmss
                     bdt.tm_year + 1900, bdt.tm_mon + 1, bdt.tm_mday,
                     bdt.tm_hour, bdt.tm_min, bdt.tm_sec);

            err = fprintf(fidw,"date: %s  %s\n",imagedate,filedate);       //  output image and file date

            if (*tags_currtags)                                            //  output filetags
               err = fprintf(fidw,"tags: %s\n",tags_currtags);
            else  err = fprintf(fidw,"tags: null" tagdelim2 "\n");         //  "null" tag if none
           
            err = fprintf(fidw,"stars: %c\n",meta_stars);                  //  output stars rating
            
            if (*meta_comments)                                            //  output user comments
               err = fprintf(fidw,"comms: %s\n",meta_comments);
            else err = fprintf(fidw,"comms: null \n");                     //  "null" if none

            if (*meta_caption)                                             //  output user caption
               err = fprintf(fidw,"capt: %s\n",meta_caption);
            else err = fprintf(fidw,"capt: null \n");                      //  "null" if none
            
            if (! *meta_city) strcpy(meta_city,"null");                    //  geotags                v.12.05
            if (! *meta_country) strcpy(meta_country,"null");
            if (! *meta_latitude) strcpy(meta_latitude,"null");
            if (! *meta_longitude) strcpy(meta_longitude,"null");
            err = fprintf(fidw,"gtags: %s^ %s^ %s^ %s \n",
                  meta_city, meta_country, meta_latitude, meta_longitude);

            err = fprintf(fidw,"\n");                                      //  EOL

            finserted = 1;
         }

         if (eof) break;
         if (fcopy) fprintf(fidw,"%s\n",buff);                             //  copy this entry
      }

      err = fclose(fidr);
      if (err) goto tagserror;
   }

   err = fclose(fidw);                                                     //  close temp file
   if (err) goto tagserror;
   
   err = rename(temp_search_index_file,search_index_file);                 //  replace tags file with temp file
   if (err) goto tagserror;
   return;
   
tagserror:
   zmessLogACK(mWin,ZTX("search index file error: %s"),strerror(errno));
   return;
}


//  file is deleted from search index file

void delete_search_index(cchar *file)
{
   char           *ppv, temp_search_index_file[200];
   char           indexbuff[indexrecl];
   char           filebuff[indexrecl], datebuff[indexrecl];
   char           tagsbuff[indexrecl], starsbuff[indexrecl];
   char           commsbuff[indexrecl], captbuff[indexrecl];
   char           gtagsbuff[indexrecl];
   int            err, ftf = 1;
   FILE           *fidr, *fidw;

   strcpy(temp_search_index_file,search_index_file);                       //  temp tag file
   strcat(temp_search_index_file,"_temp");

   fidr = fopen(search_index_file,"r");                                    //  read tag file
   if (! fidr) return;
   
   fidw = fopen(temp_search_index_file,"w");                               //  write temp tag file
   if (! fidw) goto tagserror;
   
   while (true)                                                            //  copy search index file to temp
   {                                                                       //    file, omitting this image file
      ppv = fgets_trim(indexbuff,indexrecl,fidr);

      if (! ppv || strnEqu(indexbuff,"file: ",6))                          //  EOF or start of next file
      {
         if (! ftf && strNeq(filebuff+6,file))                             //  not 1st time and not file to omit
         {
            fprintf(fidw,"%s\n",filebuff);                                 //  output completed data for previous file
            fprintf(fidw,"%s\n",datebuff);
            fprintf(fidw,"%s\n",tagsbuff);
            fprintf(fidw,"%s\n",starsbuff);
            fprintf(fidw,"%s\n",commsbuff);
            fprintf(fidw,"%s\n",captbuff);
            fprintf(fidw,"%s\n",gtagsbuff);                                //  v.12.05
            fprintf(fidw,"\n");
         }
         
         if (! ppv) break;                                                 //  EOF

         ftf = 0;
         strncpy0(filebuff,indexbuff,indexrecl);                           //  next file: filespec record
         strcpy(datebuff,"date: null ");                                   //  initz. remaining data = empty
         strcpy(tagsbuff,"tags: null"tagdelim2);
         strcpy(starsbuff,"stars: 0 ");
         strcpy(commsbuff,"comms: null ");
         strcpy(captbuff,"capt: null ");
         strcpy(gtagsbuff,"gtags: null, null, null, null ");               //  v.12.05
      }
      
      if (strnEqu(indexbuff,"date: ",6))                                   //  copy whatever is found
         strncpy0(datebuff,indexbuff,indexrecl);

      if (strnEqu(indexbuff,"tags: ",6))
         strncpy0(tagsbuff,indexbuff,indexrecl);

      if (strnEqu(indexbuff,"stars: ",7))
         strncpy0(starsbuff,indexbuff,indexrecl);

      if (strnEqu(indexbuff,"comms: ",7))
         strncpy0(commsbuff,indexbuff,indexrecl);

      if (strnEqu(indexbuff,"capt: ",6))
         strncpy0(captbuff,indexbuff,indexrecl);

      if (strnEqu(indexbuff,"gtags: ",7))                                  //  v.12.05
         strncpy0(gtagsbuff,indexbuff,indexrecl);
   }

   err = fclose(fidr);
   if (err) goto tagserror;

   err = fclose(fidw);
   if (err) goto tagserror;
   
   err = rename(temp_search_index_file,search_index_file);                 //  replace tags file with temp file
   if (err) goto tagserror;

   return;

tagserror:
   zmessLogACK(mWin,ZTX("search index file error: %s"),strerror(errno));
   return;
}


/**************************************************************************/

//  menu function - add tags to many files at once

char        **batchAddTags_filelist = 0;
int         batchAddTags_filecount = 0;
zdialog     *zdbatchAddTags = null;                                        //  batch add tags dialog

void m_batchAddTags(GtkWidget *, cchar *)
{
   void  batch_addtags_clickfunc(GtkWidget *widget, int &line, int &pos);
   void  batch_deftags_clickfunc(GtkWidget *widget, int &line, int &pos);
   int   batchAddTags_dialog_event(zdialog *zd, cchar *event);
   
   char        *ptag, **flist, *file;
   int         ii, jj, err;
   zdialog     *zd;
   GtkWidget   *widget;

   zfuncs::F1_help_topic = "batch_add_tags";
   
   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(mWin,Bexiftoolmissing);
      return;
   }

   if (mod_keep()) return;                                                 //  unsaved edits
   if (! menulock(1)) return;

   zd = zdialog_new(ZTX("Batch Add Tags"),mWin,Bproceed,Bcancel,null);
   zdbatchAddTags = zd;

   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","files","hb3",Bselectfiles,"space=10");
   zdialog_add_widget(zd,"label","labcount","hb3","0 files selected","space=10");

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labaddtags","hb1",ZTX("tags to add"),"space=10");
   zdialog_add_widget(zd,"frame","frame1","hb1",0,"expand");
   zdialog_add_widget(zd,"edit","addtags","frame1",0,"expand|wrap");

   zdialog_add_widget(zd,"hbox","space","dialog",0,"space=3");
   zdialog_add_widget(zd,"hbox","hb5","dialog");
   zdialog_add_widget(zd,"label","labdeftags","hb5",ZTX("defined tags"),"space=10");
   zdialog_add_widget(zd,"hbox","hb6","dialog",0,"expand");
   zdialog_add_widget(zd,"frame","frame5","hb6",0,"space=5|expand");
   zdialog_add_widget(zd,"scrwin","scrwin5","frame5",0,"expand");
   zdialog_add_widget(zd,"edit","deftags","scrwin5",0,"expand|wrap");

   load_deftags();                                                         //  stuff defined tags into dialog
   deftags_stuff(zd);

   batchAddTags_filelist = 0;
   batchAddTags_filecount = 0;
   *tags_batchAddTags = 0;

   widget = zdialog_widget(zd,"addtags");                                  //  tag widget mouse functions      v.12.06
   textwidget_set_clickfunc(widget,batch_addtags_clickfunc);

   widget = zdialog_widget(zd,"deftags");
   textwidget_set_clickfunc(widget,batch_deftags_clickfunc);

   zdialog_resize(zd,500,400);                                             //  run dialog
   zdialog_help(zd,"batch_add_tags");                                      //  zdialog help topic
   zdialog_run(zd,batchAddTags_dialog_event);
   
   zdialog_wait(zd);                                                       //  wait for dialog completion

   flist = batchAddTags_filelist;                                          //  selected files

   if (zd->zstat != 1) goto cleanup;
   if (! batchAddTags_filecount) goto cleanup;
   if (*tags_batchAddTags <= ' ') goto cleanup;

   write_popup_text("open","Adding Tags",500,200,mWin);                    //  status monitor popup window

   for (ii = 0; flist[ii]; ii++)                                           //  loop all selected files
   {
      file = flist[ii];                                                    //  display image
      err = f_open(file,0);
      if (err) continue;

      write_popup_text("write",file);                                      //  report progress
      zmainloop();

      load_filemeta(file);                                                 //  load current file tags

      for (jj = 1; ; jj++)                                                 //  add new tags unless already
      {
         ptag = (char *) strField(tags_batchAddTags,tagdelims,jj);
         if (! ptag) break;
         if (*ptag == ' ') continue;
         err = add_tag(ptag,tags_currtags,tagFcc);
         if (err == 2) 
            zmessageACK(mWin,ZTX("%s \n too many tags"),file);
      }

      save_filemeta(file);                                                 //  save tag changes
   }

   write_popup_text("write","COMPLETED");
   write_popup_text("close",0);
   
cleanup:

   zdialog_free(zdbatchAddTags);

   if (batchAddTags_filecount) {
      for (ii = 0; flist[ii]; ii++) 
         zfree(flist[ii]);
      zfree(flist);
   }

   menulock(0);
   return;
}


//  mouse click functions for widgets having tags                          //  v.12.06

void batch_addtags_clickfunc(GtkWidget *widget, int &line, int &pos)       //  a tag in the add list was clicked
{
   char     *txline, *txtag, end = 0;
   cchar    *delims = tagdelims":";
   
   txline = textwidget_get_line(widget,line);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,delims,end);
   if (! txtag) { zfree(txline); return; }
   
   del_tag(txtag,tags_batchAddTags);                                       //  remove tag from list
   zdialog_stuff(zdbatchAddTags,"addtags",tags_batchAddTags);
   
   zfree(txline);
   zfree(txtag);
   return;
}


void batch_deftags_clickfunc(GtkWidget *widget, int &line, int &pos)       //  a defined tag was clicked
{
   char     *txline, *txtag, end = 0;
   cchar    *delims = tagdelims":";
   
   txline = textwidget_get_line(widget,line);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,delims,end);
   if (! txtag) { zfree(txline); return; }
   
   add_tag(txtag,tags_batchAddTags,tagMcc);                                //  add defined tag to list
   zdialog_stuff(zdbatchAddTags,"addtags",tags_batchAddTags);
   
   zfree(txline);
   zfree(txtag);
   return;
}


//  batchAddTags dialog event function

int batchAddTags_dialog_event(zdialog *zd, cchar *event)
{
   int      ii;
   char     **flist = batchAddTags_filelist, countmess[50];
   
   if (strEqu(event,"files"))                                              //  select images to add tags
   {
      if (flist) {                                                         //  free prior list
         for (ii = 0; flist[ii]; ii++) 
            zfree(flist[ii]);
         zfree(flist);
      }

      flist = image_gallery_getfiles(0,mWin);                              //  get file list from user
      batchAddTags_filelist = flist;

      if (flist)                                                           //  count files in list
         for (ii = 0; flist[ii]; ii++);
      else ii = 0;
      batchAddTags_filecount = ii;
      
      snprintf(countmess,50,ZTX("%d files selected"),batchAddTags_filecount);
      zdialog_stuff(zd,"labcount",countmess);
   }
   
   return 1;
}
   

/**************************************************************************/

//  menu function - delete or replace a tag for many files at once

char        **batchDelTag_filelist = 0;

void m_batchDelTag(GtkWidget *, cchar *)
{
   int   batchDelTag_dialog_event(zdialog *zd, cchar *event);
   
   char           **flist, *file;
   char           deltag[tagcc], reptag[tagcc];
   int            ii, err;
   zdialog        *zd;

   zfuncs::F1_help_topic = "batch_delete_tag";
   
   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(mWin,Bexiftoolmissing);
      return;
   }

   if (mod_keep()) return;                                                 //  unsaved edits
   if (! menulock(1)) return;

   zd = zdialog_new(ZTX("Batch Delete Tag"),mWin,Bproceed,Bcancel,null);

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","lab1","hb1",ZTX("tag to remove"),"space=10");
   zdialog_add_widget(zd,"entry","deltag","hb1",0,"scc=20");

   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","lab2","hb2",ZTX("optional replacement"),"space=10");
   zdialog_add_widget(zd,"entry","reptag","hb2",0,"scc=20");

   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","files","hb3",Bselectfiles,"space=10");
   zdialog_add_widget(zd,"label","labcount","hb3",ZTX("0 files selected"),"space=10");
   
   zdialog_add_widget(zd,"hbox","hb4","dialog",0,"space=5");
   zdialog_add_widget(zd,"check","allfiles","hb4",ZTX("search all files"),"space=10");

   batchDelTag_filelist = flist = 0;

   zdialog_help(zd,"batch_delete_tag");                                    //  zdialog help topic
   zdialog_run(zd,batchDelTag_dialog_event);                               //  run dialog
   zdialog_wait(zd);                                                       //  wait for completion

   if (zd->zstat != 1) goto cleanup;                                       //  canceled

   flist = batchDelTag_filelist;                                           //  get selected files
   if (! flist) goto cleanup;
   
   zdialog_fetch(zd,"deltag",deltag,tagcc);                                //  get tag to delete
   strTrim2(deltag);                                                       //  remove leading and trailing blanks
   strToLower(deltag);                                                     //  use lower case for matching
   if (! *deltag) goto cleanup;

   zdialog_fetch(zd,"reptag",reptag,tagcc);                                //  optional replacement tag
   strTrim2(reptag);

   write_popup_text("open","Deleting Tags",500,200,mWin);                  //  status monitor popup window

   for (ii = 0; flist[ii]; ii++)                                           //  loop all selected files
   {
      file = flist[ii];
      err = f_open(file,0);                                                //  display image
      if (err) continue;

      write_popup_text("write",file);                                      //  report progress
      zmainloop();

      load_filemeta(file);                                                 //  load current file tags

      err = del_tag(deltag,tags_currtags);                                 //  remove tag, if present
      if (err) continue;                                                   //  not present

      if (*reptag) {                                                       //  add replacement tag, if defined
         err = add_tag(reptag,tags_currtags,tagFcc);
         if (err == 2) 
            zmessageACK(mWin,ZTX("%s \n too many tags"),file);
      }

      save_filemeta(file);                                                 //  save tag changes
   }
   
   write_popup_text("write","COMPLETED");
   write_popup_text("close",0);
   
cleanup:

   zdialog_free(zd);

   if (flist) {
      for (ii = 0; flist[ii]; ii++) 
         zfree(flist[ii]);
      zfree(flist);
   }

   menulock(0);
   return;
}


//  batchDelTag dialog event function

int batchDelTag_dialog_event(zdialog *zd, cchar *event)
{
   FILE           *fidr;
   struct stat    statbuf;
   char           deltag[tagcc];
   char           **flist = batchDelTag_filelist;
   int            err, nfiles, ii, allfiles;
   char           filebuff[indexrecl], tagsbuff[indexrecl];
   char           *ppv, *file, *tags, countmess[50];
   cchar          *ppf;
   int            maxfiles = 100000;                                       //  max. files with tag to delete 
   
   zdialog_fetch(zd,"deltag",deltag,tagcc);                                //  get tag to delete
   strTrim2(deltag);                                                       //  remove leading and trailing blanks
   strToLower(deltag);                                                     //  use lower case for matching

   if (zd->zstat == 1)                                                     //  [proceed]
   {
      if (! flist) {
         zmessageACK(mWin,ZTX("no files selected"));
         zd->zstat = 0;                                                    //  keep dialog active
         goto finish;
      }
      
      if (! *deltag) {
         zmessageACK(mWin,ZTX("no tag specified"));
         zd->zstat = 0;
         goto finish;
      }
      
      return 0;
   }

   if (strEqu(event,"deltag"))                                             //  tag changed
   {
      if (flist) {                                                         //  clear file list
         for (ii = 0; flist[ii]; ii++) 
            zfree(flist[ii]);
         zfree(flist);
         flist = 0;
      }

      zdialog_stuff(zd,"allfiles",0);                                      //  reset "all files" checkbox
      goto finish;
   }

   if (strEqu(event,"files"))                                              //  select images to delete tags
   {
      if (! *deltag) {
         zmessageACK(mWin,ZTX("specify tag"));
         goto finish;
      }

      if (flist) {                                                         //  free prior list
         for (ii = 0; flist[ii]; ii++) 
            zfree(flist[ii]);
         zfree(flist);
         flist = 0;
      }

      flist = image_gallery_getfiles(0,mWin);                              //  get file list from user
      zdialog_stuff(zd,"allfiles",0);                                      //  reset "all files" checkbox
      goto finish;
   }
   
   if (strEqu(event,"allfiles"))
   {
      if (flist) {                                                         //  clear file list
         for (ii = 0; flist[ii]; ii++) 
            zfree(flist[ii]);
         zfree(flist);
         flist = 0;
      }

      zdialog_fetch(zd,"allfiles",allfiles);                               //  get checkbox "all files"
      if (! allfiles) goto finish;                                         //  was unselected
      
      if (! *deltag) {
         zmessageACK(mWin,ZTX("specify tag"));
         goto finish;
      }

      fidr = fopen(search_index_file,"r");                                 //  read search index file
      if (! fidr) goto finish;                                             //  no file

      flist = (char **) zmalloc(maxfiles*sizeof(char *),"DeleteTag");      //  list for files found
      nfiles = 0;                                                          //  count files found
      flist[0] = 0;

      while (true)
      {
         ppv = fgets_trim(filebuff,indexrecl,fidr,1);                      //  next record
         if (! ppv) break;                                                 //  EOF

         if (! strnEqu(ppv,"file: ",6)) continue;                          //  file: /dir.../filename.jpg

         file = ppv+6;
         err=stat(file,&statbuf);                                          //  check file exists
         if (err) continue;
         if (! S_ISREG(statbuf.st_mode)) continue;

         ppv = fgets_trim(tagsbuff,indexrecl,fidr);                        //  next record
         if (! ppv) break;
         if (! strnEqu(ppv,"date: ",6)) continue;                          //  date: yyyy:mm:dd 

         ppv = fgets_trim(tagsbuff,indexrecl,fidr);                        //  next record
         if (! ppv) break;
         if (! strnEqu(ppv,"tags: ",6)) continue;                          //  tags: xxxx, xxxxx, ...

         tags = ppv + 6;
         strToLower(tags);                                                 //  use lower case for matching
         
         for (ii = 1; ; ii++)                                              //  step thru file tags
         {
            ppf = strField(tags,tagdelims,ii);
            if (! ppf) break;
            if (*ppf == ' ') continue;
            if (strEqu(ppf,deltag)) break;                                 //  look for my tag
         }
         
         if (! ppf) continue;                                              //  tag not found
         
         flist[nfiles] = strdupz(file,0,"batch_del_tag");
         nfiles++;
         if (nfiles == maxfiles-1) break;
      }

      fclose(fidr);                                                        //  close search index file
      flist[nfiles] = 0;                                                   //  EOL marker
   }

finish:

   batchDelTag_filelist = flist;                                           //  file list to process

   if (flist)                                                              //  count files in list
      for (ii = 0; flist[ii]; ii++);
   else ii = 0;
   
   snprintf(countmess,50,ZTX("%d files selected"),ii);                     //  stuff file count into dialog
   zdialog_stuff(zd,"labcount",countmess);
   
   return 0;
}
   

/**************************************************************************
   functions to view and edit metadata: EXIF, IPTC, etc. 
***************************************************************************/

//  menu function and popup dialog to show EXIF/IPTC data
//  window is updated when navigating to another image

void m_meta_view_short(GtkWidget *, cchar *menu)
{
   if (menu) zfuncs::F1_help_topic = "view_meta";
   meta_view(1);
   return;
}

void m_meta_view_long(GtkWidget *, cchar *menu)
{
   if (menu) zfuncs::F1_help_topic = "view_meta";
   meta_view(2);
   return;
}

void meta_view(int arg)
{
   int   meta_view_dialog_event(zdialog *zd, cchar *event);

   static int     length = 1;
   char           *buff;
   int            contx = 0;
   GtkWidget      *widget;

   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(mWin,Bexiftoolmissing);
      return;
   }
   
   if (! curr_file) return;
   
   if (arg > 0) length = arg;                                              //  change short/long report mode

   if (! zdexifview)                                                       //  popup dialog if not already
   {
      zdexifview = zdialog_new(ZTX("View Metadata"),mWin,Bcancel,null);
      zdialog_add_widget(zdexifview,"scrwin","scroll","dialog",0,"expand");
      zdialog_add_widget(zdexifview,"edit","exifdata","scroll",0,"expand|wrap");
      zdialog_resize(zdexifview,500,400);
      zdialog_help(zdexifview,"view_meta");                                //  zdialog help topic
      zdialog_run(zdexifview,meta_view_dialog_event);
   }

   widget = zdialog_widget(zdexifview,"exifdata");
   gtk_text_view_set_editable(GTK_TEXT_VIEW(widget),0);                    //  disable widget editing
   gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(widget),GTK_WRAP_NONE);       //  disable text wrap

   if (length == 1)                                                        //  short report
   {
      snprintf(command,ccc,"exiftool -s -common "                          //                            (-s v.12.06)
         "-%s -%s -%s -%s -%s -%s -%s -%s -%s -%s \"%s\" ",
         iptc_keywords_key, iptc_rating_key, exif_editlog_key, exif_comment_key, 
         iptc_caption_key, exif_focal_length_key, exif_city_key, exif_country_key, 
         exif_latitude_key, exif_longitude_key, curr_file);
   }
   else 
      snprintf(command,ccc,"exiftool -s -e \"%s\" ",curr_file);            //  long, output everything   (-s v.12.06)
      
   widget = zdialog_widget(zdexifview,"exifdata");                         //  widget for output
   wclear(widget);

   while ((buff = command_output(contx,command))) {                        //  run command, output into window
      wprintf(widget,"%s\n",buff);
      zfree(buff);                                                         //  memory leak
   }

   command_status(contx);                                                  //  free resources

   return;
}


//  dialog event and completion callback function

int meta_view_dialog_event(zdialog *zd, cchar *event)                      //  kill dialog
{
   if (! zd->zstat) return 0;
   zdialog_free(zdexifview);
   return 0;
}


/**************************************************************************/

//  edit EXIF/IPTC data - add or change specified EXIF/IPTC/etc. key

void m_meta_edit(GtkWidget *, cchar *menu)
{
   char        keyname[40], keydata[exif_maxcc];
   cchar       *pp1[1];
   char        **pp2;

   int   meta_edit_dialog_event(zdialog *zd, cchar *event);
   
   if (menu) zfuncs::F1_help_topic = "edit_meta";

   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(mWin,Bexiftoolmissing);
      return;
   }
   
   if (! curr_file) return;

   if (! zdexifedit)                                                       //  popup dialog if not already
   {   
      zdexifedit = zdialog_new(ZTX("Edit Metadata"),mWin,Bfetch,Bsave,Bcancel,null);
      zdialog_add_widget(zdexifedit,"vbox","hb1","dialog");
      zdialog_add_widget(zdexifedit,"hbox","hbkey","dialog",0,"space=5");
      zdialog_add_widget(zdexifedit,"hbox","hbdata","dialog",0,"space=5");
      zdialog_add_widget(zdexifedit,"label","labkey","hbkey","key name");
      zdialog_add_widget(zdexifedit,"entry","keyname","hbkey",0,"scc=20");
      zdialog_add_widget(zdexifedit,"label","labdata","hbdata","key value");
      zdialog_add_widget(zdexifedit,"entry","keydata","hbdata",0,"expand");
      zdialog_help(zdexifedit,"edit_meta");                                //  zdialog help topic 
      zdialog_run(zdexifedit,meta_edit_dialog_event);
   }

   zdialog_fetch(zdexifedit,"keyname",keyname,40);                         //  get key name from dialog
   strCompress(keyname);

   if (*keyname)                                                           //  update live dialog
   {
      pp1[0] = keyname;                                                    //  look for key data 
      pp2 = exif_get(curr_file,pp1,1);
      if (pp2[0]) {
         strncpy0(keydata,pp2[0],exif_maxcc);
         zfree(pp2[0]);
      }
      else *keydata = 0;
      zdialog_stuff(zdexifedit,"keydata",keydata);                         //  stuff into dialog
   }

   return;
}


//  dialog event and completion callback function

int  meta_edit_dialog_event(zdialog *zd, cchar *event)
{
   char        keyname[40], keydata[exif_maxcc];
   cchar       *pp1[1], *pp2[1];
   char        **pp3;
   int         err;
   
   if (! zd->zstat) return 0;
   
   zdialog_fetch(zd,"keyname",keyname,40);
   zdialog_fetch(zd,"keydata",keydata,exif_maxcc);
   strCompress(keyname);                                                   //  remove blanks

   if (zd->zstat == 1)                                                     //  fetch
   {
      zd->zstat = 0;
      if (! *keyname) return 0;
      pp1[0] = keyname;
      pp3 = exif_get(curr_file,pp1,1);
      if (pp3[0]) {
         strncpy0(keydata,pp3[0],exif_maxcc);
         zfree(pp3[0]);
      }
      else *keydata = 0;
      zdialog_stuff(zd,"keydata",keydata);
   }

   else if (zd->zstat == 2)                                                //  save
   {
      zd->zstat = 0;                                                       //  keep dialog active
      if (! *keyname) return 0;
      pp1[0] = keyname;
      pp2[0] = keydata;
      err = exif_put(curr_file,pp1,pp2,1);
      if (err) zmessageACK(mWin,"error: %s",strerror(err));
      if (zdexifview) meta_view(0);                                        //  update exif view if active
   }

   else zdialog_free(zdexifedit);                                          //  other

   return 1;
}


/**************************************************************************/

//  delete EXIF/IPTC data, specific key or all data

void m_meta_delete(GtkWidget *, cchar *menu)
{
   int   meta_delete_dialog_event(zdialog *zd, cchar *event);
   
   zdialog     *zd;

   zfuncs::F1_help_topic = "delete_meta";

   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(mWin,Bexiftoolmissing);
      return;
   }
   
   if (! curr_file) return;

   zd = zdialog_new(ZTX("Delete Metadata"),mWin,Bapply,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"radio","kall","hb1",ZTX("All"),"space=5");
   zdialog_add_widget(zd,"radio","key1","hb1",ZTX("One Key:"));
   zdialog_add_widget(zd,"entry","keyname","hb1",0,"scc=20");
   zdialog_stuff(zd,"key1",1);

   zdialog_help(zd,"delete_meta");                                         //  zdialog help topic 
   zdialog_run(zd,meta_delete_dialog_event);
   return;
}


//  dialog event and completion callback function

int  meta_delete_dialog_event(zdialog *zd, cchar *event)
{
   int         kall, key1, err;
   char        keyname[40];
   
   if (! zd->zstat) return 0;

   if (zd->zstat != 1) {                                                   //  canceled
      zdialog_free(zd);
      return 1;
   }
   
   zd->zstat = 0;                                                          //  dialog remains active

   zdialog_fetch(zd,"kall",kall);
   zdialog_fetch(zd,"key1",key1);
   zdialog_fetch(zd,"keyname",keyname,40);
   strCompress(keyname);

   if (kall)
      snprintf(command,ccc,"exiftool -m -q -overwrite_original -all=  \"%s\"",curr_file);
   else if (key1)
      snprintf(command,ccc,"exiftool -m -q -overwrite_original -%s=  \"%s\"",keyname,curr_file);
   else return 1;
   
   err = system(command);
   if (err) zmessageACK(mWin,"%s",wstrerror(err));
   
   if (zdexifview) meta_view(0);                                           //  update exif view if active

   return 1;
}


/**************************************************************************/

//  download world-map.jpg and cities-geotags files.
//  save in /home/<user>/.fotoxx/geotags/*

void m_download_geotags(GtkWidget *, cchar *)                              //  new v.12.05
{
   int      ii, yn, err;
   char     geotags[100], scriptfile[100];
   char     destfile1[100], sourcefile1[100];
   char     destfile2[100], sourcefile2[100];
   FILE     *fid;

   zfuncs::F1_help_topic = "download_geotags";

   if (Ngeotags) 
   {
      for (ii = 0; ii < Ngeotags; ii++)                                    //  free prior geotags memory
      {
         zfree(geocity[ii]);
         zfree(geocountry[ii]);
         zfree(geolat[ii]);
         zfree(geolong[ii]);
      }
      zfree(geocity);
      zfree(geocountry);
      zfree(geolat);
      zfree(geolong);
      Ngeotags = 0;
   }

   yn = zmessageYN(mWin,ZTX("Download geotags data (4 megabytes). \n"
                            "Save in %s/geotags/* \n"
                            "Proceed?"),get_zuserdir());
   if (! yn) return;
   
   sprintf(geotags,"%s/geotags",get_zuserdir());
   err = mkdir(geotags,0760);
   if (err && errno != EEXIST) {
      zmessageACK(mWin,"cannot create geotags directory \n %s",strerror(errno));
      return;
   }

   sprintf(sourcefile1,"http://kornelix.squarespace.com/storage/downloads/world-map.jpg");
   sprintf(destfile1,"%s/geotags/world-map.jpg",get_zuserdir());
   sprintf(sourcefile2,"http://kornelix.squarespace.com/storage/downloads/cities-geotags");
   sprintf(destfile2,"%s/geotags/cities-geotags",get_zuserdir());

   sprintf(scriptfile,"%s/geotags/download.scr",get_zuserdir());
   
   fid = fopen(scriptfile,"w");
   if (! fid) { 
      zmessageACK(mWin,"cannot create script file \n %s",strerror(errno)); 
      return;
   }
   
   fprintf(fid,"#! /bin/bash \n");                                         //  build bash script
   fprintf(fid,"#  download geotags data \n");
   fprintf(fid,"# \n");
   fprintf(fid,"echo DOWNLOAD WORLD MAP \n");
   fprintf(fid,"wget --progress=dot:mega -O %s %s 2>&1 \n",destfile1,sourcefile1);
   fprintf(fid,"err=$? \n");
   fprintf(fid,"if [ $err -ne 0 ]; then exit $err; fi \n");
   fprintf(fid,"echo DOWNLOAD GEOTAGS FILE \n");
   fprintf(fid,"wget -O %s %s 2>&1 \n",destfile2,sourcefile2);
   fprintf(fid,"err=$? \n");
   fprintf(fid,"if [ $err -ne 0 ]; then exit $err; fi \n");
   fprintf(fid,"echo DONE \n");
   fprintf(fid,"exit 0 \n");
   fclose(fid);

   chmod(scriptfile,0744);
   
   sprintf(command,"bash %s",scriptfile);                                  //  run the script
   err = popup_command(command,800,600,mWin);
   if (err) zmessageACK(mWin,ZTX("download failed"));
   
   /*  conv_geotags_file();  */                                            //  see below

   return;
}


/**************************************************************************/

//  get latitude/longitude data for a city [ country ]
//  inputs:        location[0] = city
//                 location[1] = country (opt)
//  outputs:       coord[2] = latitude -90 to +90, longitude -180 to +180
//                 matches[20][2] up to 20 matching city/country locations
//  returns:       no. matches for input city [ country ]
//                 (max. 20)
//
//  use null or "" for missing optional country input
//  coordinates are returned for the first match only, if any
//
//  initialization call: count = get_geotags(0,0,0);
//  returns the no. of city geotag records available.

int get_geotags(char *location[2], char *coord[2], char *matches[20][2])   //  new v.12.05
{
   char     geotagsfile[200], buff[200], *pp;
   int      cc, ii, Nmatch;
   float    fcoord;
   FILE     *fid;

   if (! Ngeotags)                                                         //  need to load cities-geotags file(s)
   {
      cc = maxgeotags * sizeof(char *);                                    //  allocate memory
      geocity = (char **) zmalloc(cc,"geotags");
      geocountry = (char **) zmalloc(cc,"geotags");
      geolat = (char **) zmalloc(cc,"geotags");
      geolong = (char **) zmalloc(cc,"geotags");

      for (int ff = 1; ff <= 2; ff++)                                      //  process two files:                 v.12.06
      {                                                                    //    cities-geotags-new, cities-geotags
         if (ff == 1) 
            sprintf(geotagsfile,"%s/geotags/cities-geotags-new",get_zuserdir());
         else sprintf(geotagsfile,"%s/geotags/cities-geotags",get_zuserdir());
         
         fid = fopen(geotagsfile,"r");
         if (! fid) continue;

         while (true)
         {
            pp = fgets_trim(buff,200,fid,1);                               //  read each city record into memory
            if (! pp) break;                                               //  (remove trailing \n)
            if (strlen(pp) < 2) continue;
            
            ii = Ngeotags;
            
            pp = (char *) strField(buff,'^',1);                            //  get latitude
            if (! pp) goto badrec;
            strTrim2(pp);
            geolat[ii] = strdupz(pp,0,"geotags");
            fcoord = atof(pp);
            if (fcoord < -90 || fcoord > 90) goto badrec;

            pp = (char *) strField(buff,'^',2);                            //  longitude
            if (! pp) goto badrec;
            strTrim2(pp);
            geolong[ii] = strdupz(pp,0,"geotags");
            fcoord = atof(pp);
            if (fcoord < -180 || fcoord > 180) goto badrec;

            pp = (char *) strField(buff,'^',3);                            //  city
            if (! pp) goto badrec;
            strTrim2(pp);
            geocity[ii] = strdupz(pp,0,"geotags");
               
            pp = (char *) strField(buff,'^',4);                            //  country
            if (! pp) goto badrec;
            strTrim2(pp);
            geocountry[ii] = strdupz(pp,0,"geotags");
            
            Ngeotags++;                                                    //  count good recs found
            if (Ngeotags < maxgeotags-2) continue;                         //  (-2 for 2 outer loop reps.)
            zmessageACK(mWin,"max. geotags exceeded");
            break;
         
         badrec:
            printf("bad cities-geotags record: %s \n",buff);
            continue;
         }
         
         fclose(fid);
      }
   }

   if (! Ngeotags) {
      zmessageACK(mWin,ZTX("geotags file must be downloaded"));
      return 0;
   }
   
   if (! location) return Ngeotags;                                        //  initialization-only, done

   for (ii = Nmatch  = 0; ii < Ngeotags; ii++)                             //  search for exact city [ country ]
   {
      if (strcasecmp(location[0],geocity[ii]) != 0) continue;
      if (location[1] && *location[1])
         if (strcasecmp(location[1],geocountry[ii]) != 0) continue;
      matches[Nmatch][0] = geocity[ii];
      matches[Nmatch][1] = geocountry[ii];
      Nmatch++;
      if (Nmatch == 20) break;                                             //  no more than 20 matches are reported
      if (Nmatch == 1) {
         coord[0] = geolat[ii];                                            //  return geolat/long of 1st match
         coord[1] = geolong[ii];
         StripZeros(coord[0]);
         StripZeros(coord[1]);
      }
   }
   
   if (Nmatch) return Nmatch;                                              //  exact match found

   for (ii = Nmatch  = 0; ii < Ngeotags; ii++)                             //  search for partial city [ country ]
   {
      cc = strlen(location[0]);                                            //  cc may be 0
      if (strncasecmp(location[0],geocity[ii],cc) != 0) continue;
      if (location[1] && *location[1]) {
         cc = strlen(location[1]);
         if (strncasecmp(location[1],geocountry[ii],cc) != 0) continue;
      }
      matches[Nmatch][0] = geocity[ii];
      matches[Nmatch][1] = geocountry[ii];
      Nmatch++;
      if (Nmatch == 20) break;                                             //  no more than 20 matches are reported
      if (Nmatch == 1) {
         coord[0] = geolat[ii];                                            //  return geolat/long of 1st match
         coord[1] = geolong[ii];
         StripZeros(coord[0]);
         StripZeros(coord[1]);
      }
   }

   return Nmatch;
}


/**************************************************************************/

//  Save geotags in image file EXIF.
//  If new or revised data, update cities-geotags-new file.                //  v.12.06

int put_geotags(char *location[2], char *coord[2])                         //  new v.12.05
{
   char        file1[200], file2[200], buff[200];
   cchar       *pcity, *pcountry;
   char        *pp, acoord[2][12];
   float       fcoord[2];
   int         ii, err, Fnew = 0, Fchange = 0;
   FILE        *fid1, *fid2;
   
   if (*coord[0] || *coord[1]) {                                           //  if coordinates present, validate
      if (! *coord[0] || ! *coord[1]) goto badcoord;
      fcoord[0] = atof(coord[0]);
      if (fcoord[0] < -90 || fcoord[0] > 90) goto badcoord;
      fcoord[1] = atof(coord[1]);
      if (fcoord[1] < -180 || fcoord[1] > 180) goto badcoord;
      if (fcoord[0] == 0 && fcoord[1] == 0) goto badcoord;
      sprintf(acoord[0],"%.4f",fcoord[0]);                                 //  format with std. precision
      sprintf(acoord[1],"%.4f",fcoord[1]);
      StripZeros(acoord[0]);
      StripZeros(acoord[1]);
   }
   else
      acoord[0][0] = acoord[1][0] = 0;                                     //  not present, set to ""    v.12.06

   if (strNeq(location[0],"null"))                                         //  unless null,              v.12.06.2
      *location[0] = toupper(*location[0]);                                //  force capitalization      v.12.06
   if (strNeq(location[1],"null"))
      *location[1] = toupper(*location[1]);
   
   strncpy0(meta_city,location[0],99);                                     //  save geotags in image file EXIF
   strncpy0(meta_country,location[1],99);
   strncpy0(meta_latitude,acoord[0],12);
   strncpy0(meta_longitude,acoord[1],12);
   Ftagschanged = 1;
   save_filemeta(curr_file);
   
   if (! *location[0] || ! *location[1]) return 1;                         //  quit here if data not complete
   if (! *coord[0] || ! *coord[1]) return 1;
   
   for (ii = 0; ii < Ngeotags; ii++) {                                     //  search geotags for city, country
      if (strcasecmp(location[0],geocity[ii]) != 0) continue;              //  (case-insensitive compare)
      if (strcasecmp(location[1],geocountry[ii]) == 0) break;
   }

   if (ii < Ngeotags) {                                                    //  found, check for revised lat/long
      if (strNeq(geolat[ii],acoord[0])) Fchange = 1;
      if (strNeq(geolong[ii],acoord[1])) Fchange = 1;
      if (strNeq(geocity[ii],location[0])) Fchange = 1;                    //  or revised capitalization   v.12.06
      if (strNeq(geocountry[ii],location[1])) Fchange = 1;
   }
   else Fnew = 1;                                                          //  a new city, country
   
   if (Fnew + Fchange == 0) return 1;                                      //  no change
   
   if (Ngeotags == maxgeotags) {
      zmessageACK(mWin,"max. geotags exceeded");
      return 0;
   }
   
   for (ii = Ngeotags; ii > 0; ii--) {                                     //  shift all geotag data up
      geocity[ii] = geocity[ii-1];
      geocountry[ii] = geocountry[ii-1];
      geolat[ii] = geolat[ii-1];
      geolong[ii] = geolong[ii-1];
   }
   
   geocity[0] = strdupz(location[0],0,"geotags");                          //  new/changed data is now first
   geocountry[0] = strdupz(location[1],0,"geotags");
   geolat[0] = strdupz(acoord[0],0,"geotags");
   geolong[0] = strdupz(acoord[1],0,"geotags");
   Ngeotags++;

   sprintf(file1,"%s/geotags/cities-geotags-new",get_zuserdir());          //  update file cities-geotags-new
   sprintf(file2,"%s/geotags/cities-geotags-temp",get_zuserdir());

   fid2 = fopen(file2,"w");
   if (! fid2) { 
      zmessageACK(mWin,"cannot update geotags file \n %s",strerror(errno)); 
      return 0;
   }

   snprintf(buff,200,"%s^  %s^  %s^  %s\n",                                //  new/changed data is first record
              acoord[0],acoord[1],location[0],location[1]);
   fputs(buff,fid2);
   
   fid1 = fopen(file1,"r");
   if (fid1) 
   {
      while (true)
      {
         pp = fgets_trim(buff,200,fid1,1);                                 //  read and copy geotags recs
         if (! pp) break;
         
         pcity = strField(buff,'^',3);                                     //  skip invalid recs
         pcountry = strField(buff,'^',4);
         if (! pcity || ! pcountry) continue;
         
         if (strcasecmp(pcity,location[0]) == 0 &&                         //  skip changed record
             strcasecmp(pcountry,location[1]) == 0) continue;

         fputs(buff,fid2);
         fputs("\n",fid2);
      }
     
      fclose(fid1);
   }

   err = fclose(fid2);                                                     //  file I/O error
   if (err) {
      zmessageACK(mWin,"cannot update geotags file \n %s",strerror(errno)); 
      return 0;
   }

   rename(file2,file1);                                                    //  replace modified file
   return 1;

badcoord:
   zmessageACK(mWin,ZTX("bad latitude/longitude: %s %s"),coord[0],coord[1]);
   return 0;
}


/**************************************************************************/

//  show a world map and set geotags data from mouse click on map

void get_geotags_worldmap()                                                //  new v.12.05
{
   int get_geotags_worldmap_dialog_event(zdialog *, cchar *);
   void get_geotags_worldmap_mousefunc();
   
   zdialog     *zd;
   int         nn;
   char        *curr_file_save;
   char        mapfile[200];
   PXM         *pxmtemp;
   cchar       *title = ZTX("Geotags from world map");

   nn = get_geotags(0,0,0);                                                //  initialize
   if (! nn) return;

   snprintf(mapfile,200,"%s/geotags/world-map.jpg",get_zuserdir());
   pxmtemp = PXBread(mapfile);
   if (! pxmtemp) {
      zmessageACK(mWin,"world-map.jpg is missing");
      return;
   }

   if (mod_keep()) return;                                                 //  unsaved edits
   if (! menulock(1)) return;

   curr_file_save = curr_file;                                             //  suspend current image file
   curr_file = 0;
   
   mutex_lock(&Fpixmap_lock);                                              //  main window = world map
   PXM_free(Fpxm);
   Fpxm = pxmtemp;
   Fww = Fpxm->ww;
   Fhh = Fpxm->hh;
   curr_file_bpc = 8;
   curr_file_size = 0;
   mutex_unlock(&Fpixmap_lock);

   gtk_window_set_title(MWIN,ZTX("click location on world map"));
   mwpaint2();                                                             //  repaint window

   zd = zdialog_new(title,mWin,Bapply,Bcancel,null);
   geotags_worldmap_dialog = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labcity","hb1",ZTX("city"),"space=5");
   zdialog_add_widget(zd,"entry","city","hb1",0,"expand");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labcountry","hb2",ZTX("country"),"space=5");
   zdialog_add_widget(zd,"entry","country","hb2",0,"expand");
   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","lablat","hb3","Latitude","space=3");
   zdialog_add_widget(zd,"entry","latitude","hb3",0,"scc=10");
   zdialog_add_widget(zd,"label","space","hb3",0,"space=5");
   zdialog_add_widget(zd,"label","lablong","hb3","Longitude","space=3");
   zdialog_add_widget(zd,"entry","longitude","hb3",0,"scc=10");
   zdialog_add_widget(zd,"hbox","hb4","dialog",0,"space=3");
   zdialog_add_widget(zd,"check","getcity","hb4",ZTX("use nearest city"));
   
   zdialog_stuff(zd,"getcity",1);                                          //  default, use nearest city geotags
   Fgeotags_worldmap_getcity = 1;
   takeMouse(zd,get_geotags_worldmap_mousefunc,dragcursor);                //  connect mouse 

   zdialog_run(zd,get_geotags_worldmap_dialog_event);                      //  start dialog
   zdialog_wait(zd);                                                       //  wait for completion

   freeMouse();                                                            //  restore current file
   menulock(0);
   f_open(curr_file_save,1);
   
   return;
}


//  dialog event and completion function

int get_geotags_worldmap_dialog_event(zdialog *zd, cchar *event)           //  new v.12.05
{
   zdialog_fetch(zd,"getcity",Fgeotags_worldmap_getcity);                  //  flag, use nearby city geotags
   
   if (! zd->zstat) return 0;                                              //  wait for completion

   if (zd->zstat == 1)                                                     //  apply
      Fgeotags_worldmap_data = 1;                                          //  use geotags from map
   
   zdialog_free(zd);                                                       //  kill dialog
   return 0;
}


//  Respond to mouse clicks on world map.
//  Set longitude and latitude, and city and country if very near.
//  The map is a Robinson Projection (see Wikipedia).

void get_geotags_worldmap_mousefunc()                                      //  new v.12.05
{
   void geotags_worldmap_adjust(float *glat, float *glong);

   float    PLEN[19] = { 1.0000, 0.9986, 0.9954, 0.9900, 0.9822, 0.9730, 0.9600,
                         0.9427, 0.9216, 0.8962, 0.8679, 0.8350, 0.7986, 0.7597,
                         0.7186, 0.6732, 0.6213, 0.5722, 0.5322 };
   float    PDFE[19] = { 0.0000, 0.0620, 0.1240, 0.1860, 0.2480, 0.3100, 0.3720,
                         0.4340, 0.4958, 0.5571, 0.6176, 0.6769, 0.7346, 0.7903,
                         0.8435, 0.8936, 0.9394, 0.9761, 1.0000 };

   zdialog     *zd = geotags_worldmap_dialog;
   float       longfudge = 10.1;
   int         mx, my, ii, minii;
   int         xmargin = 46;
   int         ymargin = 298;
   int         mapw = 6886;
   int         maph = 3490;
   float       flat, flong, glat, glong, plen;
   float       dist, mindist;
   
   if (! LMclick) return;
   LMclick = 0;
   
   mx = Mxclick;
   my = Myclick;
   
   mx = mx - xmargin;
   my = my - ymargin;
   if (mx < -4 || mx > mapw+3) return;
   if (my < -4 || my > maph+3) return;
   
   flat = 2.0 * my / maph - 1.0;                                           //  -1 (map top) to +1 (bottom)
   flat = fabs(flat);                                                      //  0 at equator, 1 at poles
   
   for (ii = 0; ii < 18; ii++)
      if (flat <= PDFE[ii+1]) break;
   flat = 5.0 * (ii + (flat - PDFE[ii]) / (PDFE[ii+1] - PDFE[ii]));        //  latitude 0 to 90 (N or S)

   ii = flat / 5.0;
   if (ii > 17) ii = 17;
   plen = PLEN[ii] + (flat - 5 * ii) / 5.0 * (PLEN[ii+1] - PLEN[ii]);
   flong = 180.0 * (mx - mapw / 2) / (mapw / 2 * plen);                    //  longitude -180 to +180

   if (my > maph/2) flat = -flat;                                          //  S latitude
   flong += longfudge;
   
   if (flat < -91 || flat > 91) return;                                    //  if out of bounds, do nothing
   if (flong < -181 || flong > 181) return;
   
   geotags_worldmap_adjust(&flat, &flong);                                 //  add map distortion adjustments

   if (flat < -90 || flat > 90) return;                                    //  if out of bounds, do nothing
   if (flong < -180 || flong > 180) return;

   sprintf(geotags_worldmap_clicklat,"%.4f",flat);                         //  saved clicked coordinates
   sprintf(geotags_worldmap_clicklong,"%.4f",flong);
   
   dist = mindist = 999999;
   minii = 0;

   if (Fgeotags_worldmap_getcity) 
   {
      for (ii = 0; ii < Ngeotags; ii++)                                    //  find nearest city/country
      {
         glat = atof(geolat[ii]);
         glong = atof(geolong[ii]);
         dist = (flat - glat) * (flat - glat);
         dist += (flong - glong) * (flong - glong);
         if (dist < mindist) {
            mindist = dist;
            minii = ii;
         }
      }
   }
   
   if (mindist < 1) {                                                      //  within about 80 km
      ii = minii;
      geotags_worldmap_city = geocity[ii];                                 //  use nearest city geotags data
      geotags_worldmap_country = geocountry[ii];
      geotags_worldmap_latitude = geolat[ii];
      geotags_worldmap_longitude = geolong[ii];
   }
   
   else {
      geotags_worldmap_city = 0;                                           //  city unknown
      geotags_worldmap_country = 0;
      geotags_worldmap_latitude = geotags_worldmap_clicklat;
      geotags_worldmap_longitude = geotags_worldmap_clicklong;
   }
   
   if (zd) {                                                               //  stuff dialog if present
      zdialog_stuff(zd,"city",geotags_worldmap_city);
      zdialog_stuff(zd,"country",geotags_worldmap_country);
      zdialog_stuff(zd,"latitude",geotags_worldmap_latitude);
      zdialog_stuff(zd,"longitude",geotags_worldmap_longitude);
   }
   
   return;
}


//  make latitude and longitude adjustments depending on map location
//  (compensate for map distortions)

void geotags_worldmap_adjust(float *glat, float *glong)                             //  new v.12.05
{
   float    adjust[7][2][13] = { 
  /*             -180   -150   -120   -90    -60    -30     0      30     60     90     120    150    180     */
  /*  90 */  { {  0,     0,     0,     0.1,  -0.1,   0.4,   0.5,   0.4,   0.6,   0.6,   0.7,   0.7,   0.7  },
               {  1.8,   1.8,   1.4,   1.1,   1.1,   0.7,   0.7,   0.7,   0.9,   1.1,   1.4,   1.6,   1.5  } },
  /*  60 */  { { -0.5,  -0.5,  -0.3,  -0.3,  -0.16, -0.16, -0.06, -0.06,  0.1,   0.1,   0.14,  0.05,  0.24 },
               {  1.1,   1.1,   0.7,   0.6,   0.4,   0.3,   0.4,   0.4,   0.4,   0.4,   0.5,   0.4,   0.7  } },
  /*  30 */  { { -0.32, -0.32, -0.27, -0.2,  -0.2,  -0.13, -0.04, -0.04,  0.05,  0.1,   0.1,   0.2,   0.24 },
               {  0.3,   0.3,   0.25,  0.13,  0.1,   0.05,  0.08,  0.1,   0.14,  0.14,  0.15,  0.13,  0.16 } },
  /*  0  */  { { -0.23, -0.23, -0.23, -0.18, -0.14, -0.09, -0.04,  0,     0.04,  0.09,  0.14,  0.18,  0.23 },
               {  0,     0,     0,    -0.1,  -0.1,  -0.1,  -0.06,  0,     0,     0,     0,    -0.06, -0.12 } },
  /* -30 */  { { -0.1,  -0.1,  -0.1,  -0.1,  -0.05, -0.05,  0,     0.08,  0.08,  0.13,  0.13,  0.18,  0.18 },
               { -0.2,  -0.2,  -0.2,  -0.25, -0.3,  -0.25, -0.2,  -0.1,  -0.16, -0.2,  -0.24, -0.3,  -0.32 } },
  /* -60 */  { {  0,     0,    -0.04, -0.05, -0.05,  0,     0,     0.06,  0.06,  0.1,   0.16,  0.1,   0.21 },
               { -0.36, -0.36, -0.5,  -0.5,  -0.4,  -0.5,  -0.35, -0.3,  -0.3,  -0.44, -0.45, -0.5,  -0.8  } },
  /* -90 */  { {  0.1,   0.1,   0.1,   0,     0,     0,     0.1,   0.05,  0.1,   0.1,   0.1,   0.1,   0.1  },
               {  0,     0,    -0.2,  -0.4,  -0.5,  -0.6,  -0.6,  -0.6,  -0.7,  -0.9,  -0.9,  -1.1,  -1.1  } } };

   int      ii, jj, kk;
   int      corner[4][2];
   float    dist[4], weight[4];
   float    s1, s2, sumweight;

   ii = (90 - *glat) / 30;                                                 //  glat 90 to -90 >> ii 0 to 6
   jj = (180 + *glong) / 30;                                               //  glong -180 to +180 >> jj 0 to 12
   
   if (ii < 0) ii = 0;                                                     //  keep 4 corners below within 
   if (ii > 5) ii = 5;                                                     //   +-90 and +-180
   if (jj < 0) jj = 0;
   if (jj > 11) jj = 11;
   
   corner[0][0] = 90 - 30 * ii;                                            //  upper left corner of 30x30 degree area
   corner[0][1] = 30 * jj - 180;
   corner[1][0] = 90 - 30 * ii;                                            //  upper right
   corner[1][1] = 30 * jj - 150;
   corner[2][0] = 60 - 30 * ii;                                            //  lower right
   corner[2][1] = 30 * jj - 150;
   corner[3][0] = 60 - 30 * ii;                                            //  lower left
   corner[3][1] = 30 * jj - 180;
   
   for (kk = 0; kk < 4; kk++)                                              //  4 corners of 30x30 area 
   {                                                                       //    containing (glat,glong)
      s1 = (*glat - corner[kk][0]);
      s2 = (*glong - corner[kk][1]);
      dist[kk] = sqrt(s1 * s1 + s2 * s2);                                  //  distance of (glat,glong) from corner kk
      weight[kk] = 1.0 - dist[kk] / 30.0;                                  //  corner weight
      if (weight[kk] < 0) weight[kk] = 0;
   }
   
   sumweight = weight[0] + weight[1] + weight[2] + weight[3];
   for (kk = 0; kk < 4; kk++)                                              //  make weights add to 1.0
      weight[kk] = weight[kk] / sumweight;
   
   *glat += weight[0] * adjust[ii][0][jj];                                 //  add weighted adjustments to (glat,glong)
   *glong += weight[0] * adjust[ii][1][jj];
   *glat += weight[1] * adjust[ii][0][jj+1];
   *glong += weight[1] * adjust[ii][1][jj+1];
   *glat += weight[2] * adjust[ii+1][0][jj+1];
   *glong += weight[2] * adjust[ii+1][1][jj+1];
   *glat += weight[3] * adjust[ii+1][0][jj];
   *glong += weight[3] * adjust[ii+1][1][jj];

   return;
}


/**************************************************************************/

//  add geotags to an image - city, country, latitude, longitude

void m_edit_geotags(GtkWidget *, cchar *menu)                              //  new v.12.05
{
   int  edit_geotags_dialog_event(zdialog *zd, cchar *event);

/**
                  Edit Geotags

      city [______________]  country [______________]
      latitude [_______] longitude [_______]
      
      [find] [web] [prev] [map] [apply] [clear] [cancel]

**/

   cchar    *title = ZTX("Edit Geotags");
   zdialog  *zd;
   int      nn;

   if (menu) zfuncs::F1_help_topic = "edit_geotags";

   if (! curr_file) return;

   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(mWin,Bexiftoolmissing);
      return;
   }

   nn = get_geotags(null,null,null);                                       //  initialize
   if (! nn) return;

   if (! zdeditgeotags)                                                    //  start dialog if not already
   {   
      zdeditgeotags = zdialog_new(title,mWin,Bfind,Bweb,Bprev,Bmap,Bapply,Bclear,Bcancel,null);
      zd = zdeditgeotags;
      zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
      zdialog_add_widget(zd,"label","labcity","hb1",ZTX("city"),"space=5");
      zdialog_add_widget(zd,"entry","city","hb1",0,"expand");
      zdialog_add_widget(zd,"label","space","hb1",0,"space=5");
      zdialog_add_widget(zd,"label","labcountry","hb1",ZTX("country"),"space=5");
      zdialog_add_widget(zd,"entry","country","hb1",0,"expand");
      zdialog_add_widget(zd,"hbox","hb4","dialog",0,"space=3");
      zdialog_add_widget(zd,"label","lablat","hb4","Latitude","space=3");
      zdialog_add_widget(zd,"entry","latitude","hb4",0,"scc=10");
      zdialog_add_widget(zd,"label","space","hb4",0,"space=5");
      zdialog_add_widget(zd,"label","lablong","hb4","Longitude","space=3");
      zdialog_add_widget(zd,"entry","longitude","hb4",0,"scc=10");

      zdialog_help(zd,"edit_geotags");                                     //  zdialog help topic
      zdialog_run(zd,edit_geotags_dialog_event);
   }
   
   load_filemeta(curr_file);                                               //  get current image geotags (EXIF)
   zd = zdeditgeotags;
   zdialog_stuff(zd,"city",meta_city);
   zdialog_stuff(zd,"country",meta_country);
   zdialog_stuff(zd,"latitude",meta_latitude);
   zdialog_stuff(zd,"longitude",meta_longitude);

   return;
}


//  dialog event and completion callback function

int edit_geotags_dialog_event(zdialog *zd, cchar *event)                   //  new v.12.05
{
   cchar * web_geocode(char *location[2], char *coord[2]);

   int          zstat, Nmatch;
   char         *location[2], *coord[2], *matches[20][2];
   char         city[100], country[100];
   char         latitude[20], longitude[20];
   float        fcoord[2];
   cchar        *errmess;
   static char  prevcity[100] = "", prevcountry[100] = "";
   static char  prevlatitude[100] = "", prevlongitude[100] = "";

   if (! zd->zstat) return 1;                                              //  wait for action button
   zstat = zd->zstat;
   zd->zstat = 0;                                                          //  keep dialog active

   zdialog_fetch(zd,"city",city,99);                                       //  get city [country] from dialog
   zdialog_fetch(zd,"country",country,99);
   location[0] = city;
   location[1] = country;
   zdialog_fetch(zd,"latitude",latitude,12);                               //  and latitude, longitude
   zdialog_fetch(zd,"longitude",longitude,12);
   coord[0] = latitude;
   coord[1] = longitude;

   if (zstat == 1)                                                         //  [find]
   {
      Nmatch = get_geotags(location,coord,matches);                        //  find in city-geotags file
      if (Nmatch == 0)                                                     //  no matches
         zmessageACK(mWin,ZTX("city not found"));

      else if (Nmatch == 1) {                                              //  one match
         zdialog_stuff(zd,"city",matches[0][0]);                           //  stuff matching city data into dialog
         zdialog_stuff(zd,"country",matches[0][1]);
         zdialog_stuff(zd,"latitude",coord[0]);
         zdialog_stuff(zd,"longitude",coord[1]);
      }

      else {                                                               //  multiple matching cities
         zstat = geotags_choosecity(location,coord);                       //  ask user to choose one
         if (zstat == 1) {                                                 //  response is available
            zdialog_stuff(zd,"city",location[0]);                          //  stuff matching city data into dialog
            zdialog_stuff(zd,"country",location[1]);
            zdialog_stuff(zd,"latitude",coord[0]);
            zdialog_stuff(zd,"longitude",coord[1]);
         }
      }
   }

   else if (zstat == 2)                                                    //  [web]
   {
      errmess = web_geocode(location,coord);                               //  look-up in web service
      if (errmess) 
         zmessageACK(mWin,errmess);                                        //  fail
      else {
         zdialog_stuff(zd,"city",location[0]);                             //  success, return all data
         zdialog_stuff(zd,"country",location[1]);                          //  (location may have been completed)
         zdialog_stuff(zd,"latitude",coord[0]);
         zdialog_stuff(zd,"longitude",coord[1]);
      }
   }

   else if (zstat == 3)                                                    //  [prev] 
   {
      zdialog_stuff(zd,"city",prevcity);                                   //  get last-used geotags
      zdialog_stuff(zd,"country",prevcountry);
      zdialog_stuff(zd,"latitude",prevlatitude);
      zdialog_stuff(zd,"longitude",prevlongitude);
   }

   else if (zstat == 4)                                                    //  [map]
   {
      zdialog_show(zd,0);
      get_geotags_worldmap();                                              //  get geotags from world map
      zdialog_show(zd,1);

      if (Fgeotags_worldmap_data) {
         zdialog_stuff(zd,"city",geotags_worldmap_city);
         zdialog_stuff(zd,"country",geotags_worldmap_country);
         zdialog_stuff(zd,"latitude",geotags_worldmap_latitude);
         zdialog_stuff(zd,"longitude",geotags_worldmap_longitude);
         Fgeotags_worldmap_data = 0;
      }
   }
   
   else if (zstat == 5)                                                    //  [apply]
   {
      if (strEqu(coord[0],"null")) *coord[0] = 0;                          //  replace "null" with ""    v.12.06.2
      if (strEqu(coord[1],"null")) *coord[1] = 0;

      if (*coord[0] || *coord[1]) {                                        //  if coordinates present, validate
         if (! *coord[0] || ! *coord[1]) goto badcoord;
         fcoord[0] = atof(coord[0]);
         if (fcoord[0] < -90 || fcoord[0] > 90) goto badcoord;
         fcoord[1] = atof(coord[1]);
         if (fcoord[1] < -180 || fcoord[1] > 180) goto badcoord;
         if (fcoord[0] == 0 && fcoord[1] == 0) goto badcoord;
      }

      put_geotags(location,coord);                                         //  update EXIF and geotags city data
      strcpy(prevcity,city);                                               //  save data for later use as [prev]
      strcpy(prevcountry,country);
      strcpy(prevlatitude,latitude);
      strcpy(prevlongitude,longitude);
   }

   else if (zstat == 6)                                                    //  [clear]
   {
      zdialog_stuff(zd,"city","");                                         //  erase dialog fields
      zdialog_stuff(zd,"country","");
      zdialog_stuff(zd,"latitude","");
      zdialog_stuff(zd,"longitude","");
   }

   else {                                                                  //  cancel
      zdialog_free(zd);
      zdeditgeotags = 0;
   }

   return 1;

badcoord:
   zmessageACK(mWin,ZTX("bad latitude/longitude: %s %s"),coord[0],coord[1]);
   return 1;
}


//  dialog to choose one city from multiple options
//  location[2] is input city and optional country
//  (may be substrings, may have multiple matches in city geotags data)
//  location[2] is output unique city and country after user choice
//  coord[2] is output latitude, longitude

char     *geotags_chosenlocation[2];
char     *geotags_chosencoord[2];

int geotags_choosecity(char *location[2], char *coord[2])                  //  new v.12.05
{
   int  geotags_choosecity_event(zdialog *zd, cchar *event);

   char     *matches[20][2], text[200];
   int      Nmatch, ii, zstat;
   zdialog  *zd;

   Nmatch = get_geotags(location,coord,matches);                           //  get matching city geotags data

   if (Nmatch == 0) return 0;                                              //  no match

   if (Nmatch == 1) {                                                      //  one match, done
      location[0] = matches[0][0];
      location[1] = matches[0][1];
      return 1;
   }

   zd = zdialog_new(ZTX("choose city"),mWin,BOK,Bcancel,null);             //  multiple matches, start dialog
   zdialog_add_widget(zd,"comboE","cities","dialog",0,"space=5");
   for (ii = 0; ii < Nmatch; ii++) {                                       //  list matching cities to choose from
      snprintf(text,200,"%s | %s",matches[ii][0],matches[ii][1]);
      zdialog_cb_app(zd,"cities",text);                                    //  duplicates are removed
   }

   zdialog_resize(zd,300,100);
   zdialog_run(zd,geotags_choosecity_event);                               //  run dialog, wait for completion
   zstat = zdialog_wait(zd);
   zdialog_free(zd);

   if (zstat == 1) {                                                       //  valid response available
      location[0] = geotags_chosenlocation[0];
      location[1] = geotags_chosenlocation[1];
      coord[0] = geotags_chosencoord[0];
      coord[1] = geotags_chosencoord[1];
      return 1;
   }
   
   return 0;
}


//  dialog event function - get chosen city/country from multiple choices

int geotags_choosecity_event(zdialog *zd, cchar *event)
{
   char           text[200];
   static char    city[100], country[100];
   char           *location[2], *coord[2];
   char           *matches[20][2];
   cchar          *pp;
   int            nn;
   static int     ftf = 1;
   
   if (ftf) {
      zdialog_cb_popup(zd,"cities");                                       //  first time, open combo box list
      ftf = 0;
   }

   if (strEqu(event,"cities")) {                                           //  OK
      zdialog_fetch(zd,"cities",text,200);

      pp = strField(text,'|',1);
      if (pp) strncpy0(city,pp,99);
      pp = strField(text,'|',2);
      if (pp) strncpy0(country,pp,99);
      strTrim2(city);
      strTrim2(country);
      location[0] = city;
      location[1] = country;

      nn = get_geotags(location,coord,matches);                            //  find in city geotags data
      if (nn) {
         geotags_chosenlocation[0] = location[0];                          //  use 1st match if > 1
         geotags_chosenlocation[1] = location[1];
         geotags_chosencoord[0] = coord[0];
         geotags_chosencoord[1] = coord[1];
         zd->zstat = 1;
      }
      else zd->zstat = 2;                                                  //  bad status
   }

   if (zd->zstat) ftf = 1;

   return 1;
}


/**************************************************************************/

//  convert a city [country] to latitude/longitude using Yahoo web service
//  (incomplete names may be completed with a very bad guess)

cchar * web_geocode(char *location[2], char *coord[2])
{
   int            err, cc;
   static char    city[100], country[100], latitude[20], longitude[20];
   char           outfile[100], URI[200], request[300];
   char           *pp1, *pp2, buffer[1000];
   FILE           *fid;
   cchar          *notfound = ZTX("not found");
   
   *coord[0] = *coord[1] = 0;                                              //  null outputs
   *city = *country = *latitude = *longitude = 0;
   
   snprintf(outfile,100,"%s/geotags/web-data",get_zuserdir());
   snprintf(URI,199,"\"http://where.yahooapis.com/geocode?q=%s,%s\"",location[0],location[1]);
   snprintf(request,299,"wget -o /dev/null -O %s %s",outfile,URI);

   err = system(request);                                                  //  send web query
   if (err) return wstrerror(err);
   
   fid = fopen(outfile,"r");                                               //  get response
   if (! fid) return notfound;
   
   while (true)
   {
      pp1 = fgets(buffer,999,fid);
      if (! pp1) break;

      pp1 = strstr(buffer,"<line2>");                                      //  parse <line2> ... </line2>
      if (pp1) {
         pp2 = strstr(pp1+7,"</line2>");
         if (pp2) {
            cc = pp2 - pp1 - 7;
            if (cc < 99) strncpy0(city,pp1+7,cc+1);
         }
      }

      pp1 = strstr(buffer,"<line4>");                                      //  parse <line4> ... </line4>
      if (pp1) {
         pp2 = strstr(pp1+7,"</line4>");
         if (pp2) {
            cc = pp2 - pp1 - 7;
            if (cc < 99) strncpy0(country,pp1+7,cc+1);
         }
      }

      pp1 = strstr(buffer,"<latitude>");                                   //  parse <latitude> ... </latitude>
      if (pp1) {
         pp2 = strstr(pp1+10,"</latitude>");
         if (pp2) {
            cc = pp2 - pp1 - 10;
            if (cc > 12) cc = 12;
            strncpy0(latitude,pp1+10,cc+1);
         }
      }

      pp1 = strstr(buffer,"<longitude>");                                  //  parse <longitude> ... </longitude>
      if (pp1) {
         pp2 = strstr(pp1+10,"</longitude>");
         if (pp2) {
            cc = pp2 - pp1 - 11;
            if (cc > 12) cc = 12;
            strncpy0(longitude,pp1+11,cc+1);
         }
      }
   }
   
   fclose(fid);
   
   if (! *city || ! *country || ! *latitude || ! *longitude)               //  check completeness of response
      return notfound;

   if (*latitude == '0' && *longitude == '0')                              //  disallow coordinate (0,0)
      return notfound;

   pp1 = strchr(latitude,'.');                                             //  keep max. 4 decimal digits
   if (pp1) *(pp1+5) = 0;
   pp1 = strchr(longitude,'.');
   if (pp1) *(pp1+5) = 0;
   
   location[0] = city;                                                     //  return all data
   location[1] = country;
   coord[0] = latitude;
   coord[1] = longitude;

   return 0;
}


/**************************************************************************/

//  batch add geotags - set geotags for multiple image files

char     **batch_geotags_filelist = 0;
int      batch_geotags_filecount = 0;

void m_batch_geotags(GtkWidget *, cchar *menu)                             //  new v.12.05
{
   int   batch_geotags_dialog_event(zdialog *zd, cchar *event);

   cchar       *title = ZTX("Batch Add Geotags");
   char        *file, **flist;
   zdialog     *zd;
   int         nn, ii, err;
   char        *location[2], *coord[2];
   char        city[100], country[100];
   char        latitude[20], longitude[20];

   zfuncs::F1_help_topic = "batch_add_geotags";

   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(mWin,Bexiftoolmissing);
      return;
   }

   nn = get_geotags(null,null,null);                                       //  initialize
   if (! nn) return;

   if (mod_keep()) return;                                                 //  unsaved edits
   if (! menulock(1)) return;

/**
                Batch Add Geotags

      [select files]  NN files selected
      city [______________]  country [______________]
      latitude [_______] longitude [_______]
      
      [find] [web] [proceed] [cancel]

**/

   zd = zdialog_new(title,mWin,Bfind,Bweb,Bproceed,Bcancel,null);

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","files","hb1",Bselectfiles,"space=10");
   zdialog_add_widget(zd,"label","labcount","hb1","0 files selected","space=10");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labcity","hb2",ZTX("city"),"space=5");
   zdialog_add_widget(zd,"entry","city","hb2",0,"expand");
   zdialog_add_widget(zd,"label","space","hb2",0,"space=5");
   zdialog_add_widget(zd,"label","labcountry","hb2",ZTX("country"),"space=5");
   zdialog_add_widget(zd,"entry","country","hb2",0,"expand");
   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","lablat","hb3","Latitude","space=3");
   zdialog_add_widget(zd,"entry","latitude","hb3",0,"scc=10");
   zdialog_add_widget(zd,"label","space","hb3",0,"space=5");
   zdialog_add_widget(zd,"label","lablong","hb3","Longitude","space=3");
   zdialog_add_widget(zd,"entry","longitude","hb3",0,"scc=10");

   batch_geotags_filelist = 0;
   batch_geotags_filecount = 0;
   flist = 0;

   zdialog_help(zd,"batch_add_geotags");
   zdialog_run(zd,batch_geotags_dialog_event);                             //  run dialog
   zdialog_wait(zd);                                                       //  wait for dialog completion

   if (zd->zstat != 3) goto cleanup;                                       //  status not [proceed]
   if (! batch_geotags_filecount) goto cleanup;                            //  no files selected

   zdialog_fetch(zd,"city",city,99);                                       //  get city [country] from dialog
   zdialog_fetch(zd,"country",country,99);
   location[0] = city;
   location[1] = country;
   zdialog_fetch(zd,"latitude",latitude,12);                               //  and latitude, longitude
   zdialog_fetch(zd,"longitude",longitude,12);
   coord[0] = latitude;
   coord[1] = longitude;
   
   zdialog_free(zd);

   flist = batch_geotags_filelist;                                         //  selected files
   if (! flist) goto cleanup;

   write_popup_text("open","Adding Geotags",500,200,mWin);                 //  status monitor popup window

   for (ii = 0; flist[ii]; ii++)                                           //  loop all selected files
   {
      file = flist[ii];                                                    //  display image
      err = f_open(file,0);
      if (err) continue;

      write_popup_text("write",file);                                      //  report progress
      zmainloop();

      load_filemeta(file);                                                 //  load current file tags
      put_geotags(location,coord);                                         //  update geotags part
   }

   write_popup_text("write","COMPLETED");
   write_popup_text("close",0);
   
cleanup:

   if (zd) zdialog_free(zd);

   if (flist) {
      for (ii = 0; flist[ii]; ii++) 
         zfree(flist[ii]);
      zfree(flist);
   }

   menulock(0);
   return;
}


//  batch_geotags dialog event function

int batch_geotags_dialog_event(zdialog *zd, cchar *event)
{
   int      ii, yn, zstat, Nmatch;
   char     **flist = batch_geotags_filelist;
   char     countmess[50];
   char     *location[2], *coord[2], *matches[20][2];
   char     city[100], country[100];
   char     latitude[20], longitude[20];
   cchar    *errmess;
   float    fcoord[2];

   if (strEqu(event,"files"))                                              //  select images to add tags
   {
      if (flist) {                                                         //  free prior list
         for (ii = 0; flist[ii]; ii++) 
            zfree(flist[ii]);
         zfree(flist);
      }

      flist = image_gallery_getfiles(curr_file,mWin);                      //  get file list from user
      batch_geotags_filelist = flist;

      if (flist)                                                           //  count files in list
         for (ii = 0; flist[ii]; ii++);
      else ii = 0;
      batch_geotags_filecount = ii;
      
      snprintf(countmess,50,ZTX("%d files selected"),batch_geotags_filecount);
      zdialog_stuff(zd,"labcount",countmess);

      image_gallery(0,"paint",0,m_gallery2);                               //  restore click function  bugfix v.12.06.2
   }

   if (! zd->zstat) return 1;                                              //  wait for action button
   zstat = zd->zstat;
   zd->zstat = 0;                                                          //  keep dialog active

   zdialog_fetch(zd,"city",city,99);                                       //  get city [country] from dialog
   zdialog_fetch(zd,"country",country,99);
   location[0] = city;
   location[1] = country;
   zdialog_fetch(zd,"latitude",latitude,12);                               //  and latitude, longitude
   zdialog_fetch(zd,"longitude",longitude,12);
   coord[0] = latitude;
   coord[1] = longitude;

   if (zstat == 1)                                                         //  [find]
   {
      Nmatch = get_geotags(location,coord,matches);                        //  find in city-geotags file
      if (Nmatch == 0)                                                     //  no matches
         zmessageACK(mWin,ZTX("city not found"));

      else if (Nmatch == 1) {                                              //  one match
         zdialog_stuff(zd,"city",matches[0][0]);                           //  stuff matching city data into dialog
         zdialog_stuff(zd,"country",matches[0][1]);
         zdialog_stuff(zd,"latitude",coord[0]);
         zdialog_stuff(zd,"longitude",coord[1]);
      }

      else {                                                               //  multiple matching cities
         zstat = geotags_choosecity(location,coord);                       //  ask user to choose one
         if (zstat == 1) {                                                 //  response is available
            zdialog_stuff(zd,"city",location[0]);                          //  stuff matching city data into dialog
            zdialog_stuff(zd,"country",location[1]);
            zdialog_stuff(zd,"latitude",coord[0]);
            zdialog_stuff(zd,"longitude",coord[1]);
         }
      }
   }

   else if (zstat == 2)                                                    //  [web]
   {
      errmess = web_geocode(location,coord);                               //  look-up in web service
      if (errmess) 
         zmessageACK(mWin,errmess);                                        //  fail
      else {
         zdialog_stuff(zd,"city",location[0]);                             //  success, return all data
         zdialog_stuff(zd,"country",location[1]);                          //  (location may have been completed)
         zdialog_stuff(zd,"latitude",coord[0]);
         zdialog_stuff(zd,"longitude",coord[1]);
      }
   }

   else if (zstat == 3)                                                    //  [proceed]
   {
      if (*coord[0] || *coord[1]) {                                        //  if coordinates present, validate
         if (! *coord[0] || ! *coord[1]) goto badcoord;
         fcoord[0] = atof(coord[0]);
         if (fcoord[0] < -90 || fcoord[0] > 90) goto badcoord;
         fcoord[1] = atof(coord[1]);
         if (fcoord[1] < -180 || fcoord[1] > 180) goto badcoord;
         if (fcoord[0] == 0 && fcoord[1] == 0) goto badcoord;
      }

      if (! batch_geotags_filecount) goto nofiles;

      if (! *city || ! *country || ! *latitude || ! *longitude) {          //  check data is complete
         yn = zmessageYN(mWin,ZTX("data is incomplete \n proceed?"));
         if (! yn) return 1;
      }

      zd->zstat = 3;                                                       //  OK to proceed
      zdialog_destroy(zd);
   }

   else zdialog_destroy(zd);                                               //  cancel
   return 1;

badcoord:
   zmessageACK(mWin,ZTX("bad latitude/longitude: %s %s"),coord[0],coord[1]);
   return 1;

nofiles:
   zmessageACK(mWin,ZTX("no files selected"));
   return 1;
}


/**************************************************************************/

//  Group images by geotags and date, with a count of images in each group. 
//  Click on a group to get a thumbnail gallery of all images in the group.

struct grec_t  {                                                           //  image geotags data
   char        date[12];
   char        *city, *country;
   int         count;
};
grec_t   *grec = 0;
int      Ngrec = 0;


void m_geotag_groups(GtkWidget *, cchar *)                                 //  new v.12.06
{
   int   geotag_groups_comp(cchar *rec1, cchar *rec2);
   void  geotag_groups_click(GtkWidget *widget, int &line, int &pos); 

   char           indexbuff[indexrecl];
   cchar          *pp;
   int            err, iix, iig, cc, cc1, cc2;
   FILE           *fid;
   struct stat    statbuf;
   GtkWidget      *textwin;
   char           dateg[12], cityg[100], countryg[100];
   char           datex[12], cityx[100], countryx[100];

   if (! grec) {                                                           //  allocate memory for search index data
      cc = MAXIMAGES * sizeof(grec_t);
      grec = (grec_t *) zmalloc(cc,"geotag_groups");
      memset(grec,0,cc);
   }

   for (iix = 0; iix < Ngrec; iix++) {                                     //  free previous data
      if (grec[iix].city) zfree(grec[iix].city);
      if (grec[iix].country) zfree(grec[iix].country);
      grec[iix].city = grec[iix].country = 0;
   }

   Ngrec = 0;

   fid = fopen(search_index_file,"r");
   if (! fid) {
      zmessageACK(mWin,"no index file found");
      return;
   }

   *indexbuff = 0;

   while (true)                                                            //  read index file
   {
      if (! strnEqu(indexbuff,"file: ",6)) {                               //  skip to next "file" record
         pp = fgets_trim(indexbuff,indexrecl,fid);
         if (! pp) break;                                                  //  EOF
         continue;
      }

      *indexbuff = 0;                                                      //  prevent possible looping

      err=stat(indexbuff+6,&statbuf);                                      //  check image file exists
      if (err) continue;
      if (! S_ISREG(statbuf.st_mode)) continue;                            //  no

      iix = Ngrec;                                                         //  start new grec

      strcpy(grec[iix].date,"null");                                       //  all data = null
      grec[iix].city = strdupz("null",0,"geotag_groups");
      grec[iix].country = strdupz("null",0,"geotag_groups");

      while (true)
      {
         pp = fgets_trim(indexbuff,indexrecl,fid);                         //  get recs following "file"
         if (! pp) break;
         
         if (strnEqu(indexbuff,"file: ",6)) break;                         //  ran into next file

         if (strnEqu(pp,"date: ",6)) {                                     //  get image date
            pp = strField(pp,' ',2);
            if (pp) strncpy0(grec[iix].date,pp,12);
         }
         
         if (strnEqu(pp,"gtags: ",7)) {                                    //  get image geotags
            pp = strField(indexbuff+7,'^',1);
            if (pp) {
               zfree(grec[iix].city);
               grec[iix].city = strdupz(pp,0,"geotag_groups");
            }
            pp = strField(indexbuff+7,'^',2);
            if (pp) {
               zfree(grec[iix].country);
               grec[iix].country = strdupz(pp,0,"geotag_groups");
            }
         }
      }

      if (++Ngrec == MAXIMAGES) {
         zmessageACK(mWin,"too many image files");
         return;
      }
   }

   fclose(fid);
   
   if (! Ngrec) {
      zmessageACK(mWin,"no geotags data found");
      return;
   }

   if (Ngrec > 1)                                                          //  sort index recs by date, gtags
      HeapSort((char *) grec, sizeof(grec_t), Ngrec, geotag_groups_comp);

   iig = 0;                                                                //  1st group from grec[0]
   strncpy0(dateg,grec[iig].date,12);
   strncpy0(cityg,grec[iig].city,99);
   strncpy0(countryg,grec[iig].country,99);
   grec[iig].count = 1;                                                    //  group count = 1

   for (iix = 1; iix < Ngrec; iix++)                                       //  scan following grecs
   {
      strncpy0(datex,grec[iix].date,12);
      strncpy0(cityx,grec[iix].city,99);
      strncpy0(countryx,grec[iix].country,99);

      if (strEqu(dateg,datex) && strEqu(cityg,cityx) && strEqu(countryg,countryx))
      {
         zfree(grec[iix].city);                                            //  same [date/city/country] group
         zfree(grec[iix].country);
         grec[iix].city = grec[iix].country = 0;
         grec[iig].count++;                                                //  increment group count
      }
      else 
      {
         iig++;                                                            //  new group
         if (iix > iig) {
            grec[iig] = grec[iix];                                         //  copy and pack down
            grec[iix].city = grec[iix].country = 0;
         }
         strcpy(dateg,datex);                                              //  group data
         strcpy(cityg,cityx);
         strcpy(countryg,countryx);
         grec[iig].count = 1;                                              //  group count = 1
      }
   }

   Ngrec = iig + 1;                                                        //  unique groups count

   textwin = write_popup_text("open","geotag groups",620,400,mWin);        //  write unique groups to popup window

   sprintf(indexbuff,"%-30s  %-30s  %-12s  %5s ",
               "country","city","date","count");
   write_popup_text("write",indexbuff);

   for (iig = 0; iig < Ngrec; iig++)
   {
      strncpy0(dateg,grec[iig].date,12);
      
      utf8substring(cityg,grec[iig].city,0,30);
      cc1 = 30 + strlen(cityg) - utf8len(cityg);

      utf8substring(countryg,grec[iig].country,0,30);
      cc2 = 30 + strlen(countryg) - utf8len(countryg);

      sprintf(indexbuff,"%-*s  %-*s  %-12s  %5d ",
               cc2,countryg,cc1,cityg,dateg,grec[iig].count);
      write_popup_text("write",indexbuff);
   }

   write_popup_text("top");

   textwidget_set_clickfunc(textwin,geotag_groups_click);   

   return;
}


//  Compare 2 grec records by geotags and date, 
//  return <0  0  >0   for   rec1  <  =  >  rec2.

int geotag_groups_comp(cchar *rec1, cchar *rec2)                           //  v.12.06
{
   int      ii;

   char * country1 = ((grec_t *) rec1)->country;                           //  compare countries
   char * country2 = ((grec_t *) rec2)->country;
   ii = strcmp(country1,country2);
   if (ii) return ii;

   char * city1 = ((grec_t *) rec1)->city;                                 //  compare cities
   char * city2 = ((grec_t *) rec2)->city;
   ii = strcmp(city1,city2);
   if (ii) return ii;

   char * date1 = ((grec_t *) rec1)->date;                                 //  compare dates
   char * date2 = ((grec_t *) rec2)->date;
   ii = strcmp(date1,date2);
   return ii;
}


//  Receive clicks on report window and generate gallery of images
//  matching the selected date and geotags data.

void  geotag_groups_click(GtkWidget *widget, int &line, int &pos)          //  v.12.06
{
   int      iix, match;
   cchar    *pp;
   char     date[12], city[100], country[100];
   char     imagefile[maxfcc], resultsfile[200];
   char     indexbuff[indexrecl];
   FILE     *fid1, *fid2;

   iix = line - 1;                                                         //  clicked grec[iix]
   if (iix < 0 || iix > Ngrec-1) return;

   strncpy0(date,grec[iix].date,12);
   strncpy0(city,grec[iix].city,99);
   strncpy0(country,grec[iix].country,99);

   fid1 = fopen(search_index_file,"r");                                    //  find images matching date and geotags

   snprintf(resultsfile,199,"%s/search_results",get_zuserdir());           //  output scratch file
   fid2 = fopen(resultsfile,"w");
   if (! fid2) goto writerror;

   *indexbuff = 0;

   while (true)                                                            //  read index file
   {
      if (! strnEqu(indexbuff,"file: ",6)) {                               //  skip to next "file" record
         pp = fgets_trim(indexbuff,indexrecl,fid1);
         if (! pp) break;                                                  //  EOF
         continue;
      }

      *indexbuff = 0;                                                      //  prevent possible looping

      strncpy0(imagefile,indexbuff+6,maxfcc);                              //  save filespec

      for (match = 1; ; )
      {
         pp = fgets_trim(indexbuff,indexrecl,fid1);                        //  get recs following "file"
         if (! pp) break;
         
         if (strnEqu(indexbuff,"file: ",6)) break;                         //  ran into next file

         if (strnEqu(indexbuff,"date: ",6)) {                              //  compare image date to req. date
            pp = strField(indexbuff+6,' ',1);
            if (strNeq(pp,date)) { match = 0; break; }
         }
         
         if (strnEqu(indexbuff,"gtags: ",7)) {                             //  compare image geotags to req. geotags
            pp = strField(indexbuff+7,'^',1);
            if (! pp) pp = "null";
            if (strNeq(pp,city)) { match = 0; break; }
            pp = strField(indexbuff+7,'^',2);
            if (! pp) pp = "null";
            if (strNeq(pp,country)) { match = 0; break; }
         }
      }
      
      if (match) fprintf(fid2,"%s\n",imagefile);                           //  output matching file
   }

   fclose(fid1);
   fclose(fid2);

   image_gallery(resultsfile,"initF",0,m_gallery2,mWin);                   //  generate gallery of matching files
   image_gallery(0,"paint1");                                              //  show image gallery of results
   return;
   
writerror:
   zmessLogACK(mWin,"output file error: %s",strerror(errno));
   return;
}


/**************************************************************************/

//  search image tags, dates, stars, comments, captions for matching images

namespace search_images
{
   zdialog  *zdsearchimages = 0;                                           //  search images dialog

   char     searchDateFrom[12] = "";                                       //  search images
   char     searchDateTo[12] = "";
   char     searchStarsFrom[4] = "";
   char     searchStarsTo[4] = "";

   char     searchCity[100] = "";                                          //  search geotags
   char     searchCountry[100] = "";
   char     searchLatitude[20] = "";
   char     searchLongitude[20] = "";
   float    searchKilometers = 100;
   float    flatitude = 0, flongitude = 0;

   int      Nsearchkeys = 0;
   char     *searchkeys[5];                                                //  search metadata keys and match data
   char     *searchkeydata[5];
   char     searchkeyx[8], searchkeydatax[8];

//  allocated memory for search index data

   struct xrec_t  {                                                        //  image metadata record
      char        *file;                                                   //  image filespec
      char        date[12];                                                //  image date
      char        stars;                                                   //  image star rating, '0' to '5'
      char        fill[3];                                                 //  (fill to even 4 bytes)
      char        *tags;                                                   //  image tags
      char        *comms;                                                  //  image comments
      char        *capt;                                                   //  image caption
      char        *gtags;                                                  //  image geotags data
   };
   xrec_t   *xrec = 0;
   int      Nxrec = 0;
}

using namespace search_images;


void m_search_images(GtkWidget *, cchar *)                                 //  overhauled    v.12.05
{
   void  search_searchtags_clickfunc(GtkWidget *widget, int &line, int &pos);
   void  search_deftags_clickfunc(GtkWidget *widget, int &line, int &pos);
   int   searchimages_dialog_event(zdialog*, cchar *event);

   zdialog     *zd;
   GtkWidget   *widget;

   if (mod_keep()) return;                                                 //  unsaved edits
   if (! menulock(1)) return;
   menulock(0);

   zfuncs::F1_help_topic = "search_images";

/***
                  Search Image Metadata

         date range   [_______] [_______] (yyyymmdd)
         stars range  [__] [__]                  all any
         search tags  [________________________] (o) (o)
         search text  [________________________] (o) (o)
         search files [________________________] (o) (o)
         other criteria: [geotags] [other]

         defined tags
          ----------------------------------------------
         |                                              |
         |                                              |
         |                                              |
         |                                              |
         |                                              |
          ----------------------------------------------
                                     [proceed] [cancel]
***/

   zd = zdialog_new(ZTX("Search Image Metadata"),mWin,Bproceed,Bcancel,null);
   zdsearchimages = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"homog|space=3");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"homog|space=3|expand");

   zdialog_add_widget(zd,"label","labD","vb1",ZTX("date range"));
   zdialog_add_widget(zd,"label","labS","vb1",ZTX("stars range"));
   zdialog_add_widget(zd,"label","labT","vb1",ZTX("search tags"));
   zdialog_add_widget(zd,"label","labT","vb1",ZTX("search text"));
   zdialog_add_widget(zd,"label","labF","vb1",ZTX("search files"));

   zdialog_add_widget(zd,"hbox","hbD","vb2",0,"space=1");
   zdialog_add_widget(zd,"entry","datefrom","hbD",0,"scc=10");
   zdialog_add_widget(zd,"entry","dateto","hbD",0,"scc=10");
   zdialog_add_widget(zd,"label","labD","hbD",ZTX("(yyyymmdd)"),"space=5");

   zdialog_add_widget(zd,"hbox","hbS","vb2",0,"space=1");
   zdialog_add_widget(zd,"entry","starsfrom","hbS",0,"scc=2");
   zdialog_add_widget(zd,"entry","starsto","hbS",0,"scc=2");
   zdialog_add_widget(zd,"label","space","hbS",0,"expand");
   zdialog_add_widget(zd,"label","all-any","hbS",ZTX("all/any"));

   zdialog_add_widget(zd,"hbox","hbT","vb2",0,"space=1");
   zdialog_add_widget(zd,"frame","frameT","hbT",0,"expand");
   zdialog_add_widget(zd,"edit","searchtags","frameT",0,"expand|wrap");
   zdialog_add_widget(zd,"radio","alltags","hbT",0);
   zdialog_add_widget(zd,"radio","anytags","hbT",0);

   zdialog_add_widget(zd,"hbox","hbC","vb2",0,"space=1|expand");
   zdialog_add_widget(zd,"entry","searchtext","hbC",0,"expand");
   zdialog_add_widget(zd,"radio","alltext","hbC",0);
   zdialog_add_widget(zd,"radio","anytext","hbC",0);

   zdialog_add_widget(zd,"hbox","hbF","vb2",0,"space=1|expand");
   zdialog_add_widget(zd,"entry","searchfiles","hbF",0,"expand");
   zdialog_add_widget(zd,"radio","allfiles","hbF",0);
   zdialog_add_widget(zd,"radio","anyfiles","hbF",0);
   
   zdialog_add_widget(zd,"hbox","hbO","dialog","space=3");
   zdialog_add_widget(zd,"label","labO1","hbO",ZTX("other criteria"),"space=3");
   zdialog_add_widget(zd,"button","geotags","hbO","geotags","space=8");
   zdialog_add_widget(zd,"label","geotags#","hbO","( )");
   zdialog_add_widget(zd,"label","space","hbO",0,"space=5");
   zdialog_add_widget(zd,"button","other","hbO",ZTX("other"),"space=8");
   zdialog_add_widget(zd,"label","other#","hbO","( )");
   
   zdialog_add_widget(zd,"hbox","hbA1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labdeftags","hbA1",ZTX("defined tags"),"space=5");
   zdialog_add_widget(zd,"hbox","hbA2","dialog",0,"expand");
   zdialog_add_widget(zd,"frame","frameA","hbA2",0,"space=5|expand");
   zdialog_add_widget(zd,"scrwin","scrwinA","frameA",0,"expand");
   zdialog_add_widget(zd,"edit","deftags","scrwinA",0,"expand|wrap"); 

   zdialog_stuff(zd,"datefrom",searchDateFrom);                            //  stuff previous search criteria
   zdialog_stuff(zd,"dateto",searchDateTo);
   zdialog_stuff(zd,"starsfrom",searchStarsFrom);
   zdialog_stuff(zd,"starsto",searchStarsTo);
   zdialog_stuff(zd,"searchtags",tags_searchtags);
   zdialog_stuff(zd,"searchtext",tags_searchtext);
   zdialog_stuff(zd,"searchfiles",meta_searchfiles);
   
   zdialog_stuff(zd,"alltags",0);                                          //  default any tags, text, files
   zdialog_stuff(zd,"anytags",1);
   zdialog_stuff(zd,"alltext",0);
   zdialog_stuff(zd,"anytext",1);
   zdialog_stuff(zd,"allfiles",0);
   zdialog_stuff(zd,"anyfiles",1);

   load_deftags();                                                         //  stuff defined tags into dialog
   deftags_stuff(zd);

   widget = zdialog_widget(zd,"searchtags");                               //  tag widget mouse functions      v.12.06
   textwidget_set_clickfunc(widget,search_searchtags_clickfunc);

   widget = zdialog_widget(zd,"deftags");
   textwidget_set_clickfunc(widget,search_deftags_clickfunc);

   zdialog_resize(zd,400,450);                                             //  start dialog
   zdialog_help(zd,"search_images");                                       //  zdialog help topic 
   zdialog_run(zd,searchimages_dialog_event);
   zdialog_wait(zd);                                                       //  wait for dialog completion   

   return;
}

//  mouse click functions for widgets having tags                          //  v.12.06

void search_searchtags_clickfunc(GtkWidget *widget, int &line, int &pos)   //  search tag clicked
{
   char     *txline, *txtag, end = 0;
   cchar    *delims = tagdelims":";
   
   txline = textwidget_get_line(widget,line);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,delims,end);
   if (! txtag) { zfree(txline); return; }
   
   del_tag(txtag,tags_searchtags);                                         //  remove from search list
   zdialog_stuff(zdsearchimages,"searchtags",tags_searchtags);
   
   zfree(txline);
   zfree(txtag);
   return;
}


void search_deftags_clickfunc(GtkWidget *widget, int &line, int &pos)      //  defined tag clicked
{
   char     *txline, *txtag, end = 0;
   cchar    *delims = tagdelims":";
   
   txline = textwidget_get_line(widget,line);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,delims,end);
   if (! txtag) { zfree(txline); return; }
   
   add_tag(txtag,tags_searchtags,tagScc);                                  //  add to search tag list
   zdialog_stuff(zdsearchimages,"searchtags",tags_searchtags);
   
   zfree(txline);
   zfree(txtag);
   return;
}


//  search images dialog event and completion callback function

int searchimages_dialog_event(zdialog *zd, cchar *event)                   //  overhauled    v.12.05
{
   int  searchimages_geotags_dialog();
   int  searchimages_metadata_dialog();
   int  searchimages_geotags_select(char *gtagsrec);
   int  searchimages_metadata_select(char *imagefile);
   int  searchimages_metadata_report();

   char           indexbuff[indexrecl], resultsfile[200];
   char           temp[indexrecl], date1[12], date2[12];
   cchar          *pp, *pps, *ppf;
   int            err, iix, iis, iif, cc;
   int            nmatch, nfail;
   int            Fdates, Ftext, Ffiles, Ftags, Fstars, Fgtags, Fmeta;
   int            Falltags, Falltext, Fallfiles;
   FILE           *fid;
   struct stat    statbuf;

   if (strEqu(event,"geotags"))                                            //  get geotags search criteria   v.12.05
      searchimages_geotags_dialog();

   if (strEqu(event,"other"))                                              //  get other metadata criteria   v.12.05
      searchimages_metadata_dialog();

   Fgtags = 0;
   if (*searchCity || *searchCountry ||                                    //  search geotags was given
       *searchLatitude || *searchLongitude) Fgtags = 1;
   
   Fmeta = 0;
   if (Nsearchkeys) Fmeta = 1;                                             //  search other metadata was given
   
   if (Fgtags) zdialog_stuff(zd,"geotags#","(*)");
   else zdialog_stuff(zd,"geotags#","( )");
   
   if (Fmeta) zdialog_stuff(zd,"other#","(*)");
   else zdialog_stuff(zd,"other#","( )");
   
   if (! zd->zstat) return 1;                                              //  wait for dialog completion

   if (zd->zstat != 1) {                                                   //  cancel
      zdialog_free(zd);
      return 1;
   }

   zdialog_fetch(zd,"datefrom",searchDateFrom,10);                         //  get search date range
   zdialog_fetch(zd,"dateto",searchDateTo,10);
   zdialog_fetch(zd,"starsfrom",searchStarsFrom,2);                        //  get search stars range
   zdialog_fetch(zd,"starsto",searchStarsTo,2);
   zdialog_fetch(zd,"searchtags",tags_searchtags,tagScc);                  //  get search tags
   zdialog_fetch(zd,"searchtext",tags_searchtext,tagScc);                  //  get search text*
   zdialog_fetch(zd,"searchfiles",meta_searchfiles,tagScc);                //  get search /path*/file*

   zdialog_fetch(zd,"alltags",Falltags);                                   //  get match all/any options
   zdialog_fetch(zd,"alltext",Falltext);
   zdialog_fetch(zd,"allfiles",Fallfiles);

   zdialog_free(zd);                                                       //  cancel dialog

   strcpy(date1,"0000:01:01");                                             //  defaults for missing dates
   strcpy(date2,"9999:12:31"); 

   Fdates = 0;
   if (*searchDateFrom) {                                                  //  date from was given
      Fdates++;
      strncpy(date1,searchDateFrom,4);                                     //  convert format
      if (searchDateFrom[4] >= '0')                                        //    yyyymmdd >> yyyy:mm:dd
         strncpy(date1+5,searchDateFrom+4,2);
      if (searchDateFrom[6] >= '0')
         strncpy(date1+8,searchDateFrom+6,2);
   }
   if (*searchDateTo) {                                                    //  date to was given
      Fdates++;
      strncpy(date2,searchDateTo,4);
      if (searchDateTo[4] >= '0')
         strncpy(date2+5,searchDateTo+4,2);
      if (searchDateTo[6] >= '0')
         strncpy(date2+8,searchDateTo+6,2);
   }

   Fstars = 0;
   if (*searchStarsFrom || *searchStarsTo) Fstars = 1;                     //  stars was given
   
   Ffiles = 0;
   if (! blank_null(meta_searchfiles)) Ffiles = 1;                         //  search path / file (fragment) was given

   Ftext = 0;
   if (! blank_null(tags_searchtext)) Ftext = 1;                           //  search text was given

   Ftags = 0;
   if (! blank_null(tags_searchtags)) Ftags = 1;                           //  search tags was given
   
   if (Ffiles) strToLower(meta_searchfiles);                               //  all comparisons in lower case
   if (Ftags) strToLower(tags_searchtags);
   if (Ftext) strToLower(tags_searchtext);

   if (! xrec) {                                                           //  allocate memory for search index data
      cc = MAXIMAGES * sizeof(xrec_t);
      xrec = (xrec_t *) zmalloc(cc,"search_images");
      memset(xrec,0,cc);
   }

   for (iix = 0; iix < Nxrec; iix++) {                                     //  free previous data
      if (xrec[iix].file) zfree(xrec[iix].file);
      if (xrec[iix].tags) zfree(xrec[iix].tags);
      if (xrec[iix].comms) zfree(xrec[iix].comms);
      if (xrec[iix].capt) zfree(xrec[iix].capt);
      if (xrec[iix].gtags) zfree(xrec[iix].gtags);
      xrec[iix].file = 0;                                                  //  v.12.06
      xrec[iix].tags = 0;
      xrec[iix].comms = 0;
      xrec[iix].capt = 0;
      xrec[iix].gtags = 0;
   }

   Nxrec = 0;

   fid = fopen(search_index_file,"r");                                     //  open index file
   if (! fid) goto noindexfile;
   *indexbuff = 0;

   while (true)                                                            //  read index file
   {
      if (! strnEqu(indexbuff,"file: ",6)) {                               //  skip to next "file" record
         pp = fgets_trim(indexbuff,indexrecl,fid);
         if (! pp) break;                                                  //  EOF
         continue;
      }

      *indexbuff = 0;                                                      //  prevent possible looping

      iix = Nxrec;
      if (xrec[iix].file) zfree(xrec[iix].file);                           //  xrecs are re-used in loop
      if (xrec[iix].tags) zfree(xrec[iix].tags);
      if (xrec[iix].comms) zfree(xrec[iix].comms);
      if (xrec[iix].capt) zfree(xrec[iix].capt);
      if (xrec[iix].gtags) zfree(xrec[iix].gtags);
      xrec[iix].file = 0;                                                  //  v.12.06
      xrec[iix].tags = 0;
      xrec[iix].comms = 0;
      xrec[iix].capt = 0;
      xrec[iix].gtags = 0;

      memset(&xrec[iix],0,sizeof(xrec));
      pp = indexbuff + 6;                                                  //  filespec
      xrec[iix].file = strdupz(pp,0,"search_images");                      //  setup new xrec member
      xrec[iix].stars = '0';

      while (true)
      {
         pp = fgets_trim(indexbuff,indexrecl,fid);                         //  get recs following "file"
         if (! pp) break;

         if (strnEqu(pp,"date: ",6))
         {
            pp = strField(pp,' ',2);
            if (pp) strncpy0(xrec[iix].date,pp,12);
         }
         
         else if (strnEqu(pp,"tags: ",6)) 
            xrec[iix].tags = strdupz(pp+6,0,"search_images");

         else if (strnEqu(pp,"stars: ",7))
            xrec[iix].stars = *(pp+7);

         else if (strnEqu(pp,"comms: ",7))
            xrec[iix].comms = strdupz(pp+7,0,"search_images");

         else if (strnEqu(pp,"capt: ",6))
            xrec[iix].capt = strdupz(pp+6,0,"search_images");

         else if (strnEqu(pp,"gtags: ",7))
            xrec[iix].gtags = strdupz(pp+7,0,"search_images");
         
         else break;
      }

      err=stat(xrec[iix].file,&statbuf);                                   //  check file exists
      if (err) continue;
      if (! S_ISREG(statbuf.st_mode)) continue;

      if (Ffiles)                                                          //  file name match is wanted
      {
         nmatch = nfail = 0;

         for (iis = 1; ; iis++)
         {
            pps = strField(meta_searchfiles,' ',iis);                      //  step thru search file names
            if (! pps) break;
            if (strcasestr(xrec[iix].file,pps)) nmatch++;
            else nfail++;
         }

         if (nmatch == 0) continue;                                        //  no match to any file
         if (Fallfiles && nfail) continue;                                 //  no match to all files 
      }

      if (Fdates)                                                          //  date match is wanted
      {
         if (strcmp(xrec[iix].date,date1) < 0) continue;
         if (strcmp(xrec[iix].date,date2) > 0) continue;
      }

      if (Ftags)                                                           //  tags match is wanted
      {
         strToLower(temp,xrec[iix].tags);
         nmatch = nfail = 0;

         for (iis = 1; ; iis++)                                            //  step thru search tags
         {
            pps = strField(tags_searchtags,tagdelims,iis);                 //  (delimited, wildcards)
            if (! pps) break;
            if (*pps == ' ') continue;

            for (iif = 1; ; iif++)                                         //  step thru file tags (delimited)
            {
               ppf = strField(temp,tagdelims,iif);
               if (! ppf) { nfail++; break; }                              //  count matches and fails
               if (*ppf == ' ') continue;
               if (MatchWild(pps,ppf) == 0) { nmatch++; break; }           //  wildcard match 
            }
         }

         if (nmatch == 0) continue;                                        //  no match to any tag
         if (Falltags && nfail) continue;                                  //  no match to all tags
      }

      if (Fstars)                                                          //  stars match is wanted
      {
         if (*searchStarsFrom && xrec[iix].stars < *searchStarsFrom) continue;
         if (*searchStarsTo && xrec[iix].stars > *searchStarsTo) continue;
      }

      if (Ftext)                                                           //  text match is wanted
      {
         nmatch = nfail = 0;

         for (iis = 1; ; iis++)                                            //  step through search words
         {
            pps = strField(tags_searchtext,' ',iis);
            if (! pps) break;
            if (*pps == ' ') continue;
            if (strcasestr(xrec[iix].capt,pps)) nmatch++;                  //  search captions for word
            else if (strcasestr(xrec[iix].comms,pps)) nmatch++;            //  search comments for word
            else nfail++;
         }
                     
         if (nmatch == 0) continue;                                        //  no match to any word
         if (Falltext && nfail) continue;                                  //  no match to all words
      }

      if (Fgtags)                                                          //  geotags match is wanted   v.12.05
         if (! searchimages_geotags_select(xrec[iix].gtags)) continue;
      
      if (Fmeta)                                                           //  other metadata match      v.12.05
         if (! searchimages_metadata_select(xrec[iix].file)) continue;

      if (++Nxrec == MAXIMAGES) goto toomanyfiles;                         //  keep this record
   }

   fclose(fid);

   if (! Nxrec) goto nomatches;
   
   snprintf(resultsfile,199,"%s/search_results",get_zuserdir());
   fid = fopen(resultsfile,"w");                                           //  write search results to file
   if (! fid) goto writerror;
   
   for (iix = 0; iix < Nxrec; iix++)                                       //  write matching filespecs
      fprintf(fid,"%s\n",xrec[iix].file);

   fclose(fid);

   image_gallery(resultsfile,"initF",0,m_gallery2,mWin);                   //  generate gallery of matching files

   if (Fmeta)                                                              //  metadata report format    v.12.05
      searchimages_metadata_report();
   else {
      image_gallery(0,"paint1");                                           //  show image gallery of results
   }

   return 0;
   
nomatches:
   zmessageACK(mWin,ZTX("No matching images found"));
   return 0;

noindexfile:
   zmessageACK(mWin,ZTX("No search index file present"));
   return 0;

writerror:
   zmessLogACK(mWin,"output file error: %s",strerror(errno));
   return 0;

toomanyfiles:
   zmessLogACK(mWin,"too many image files");
   return 0;
}


/**************************************************************************/

//  dialog to get geotags search criteria

int searchimages_geotags_dialog()                                          //  v.12.05
{
   int searchimages_geotags_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   int         zstat;

/***
            Add Geotags Search Criteria

         city [__________]  country [__________]
         latitude [_______]  longitude [_______] 
         [find]  [map]  KM range [___]
         
                             [apply]  [cancel]

***/

   zd = zdialog_new(ZTX("Add Geotags Search Criteria"),mWin,Bapply,Bcancel,null);

   zdialog_add_widget(zd,"hbox","hbg1","dialog",0);
   zdialog_add_widget(zd,"label","labcity","hbg1",ZTX("city"),"space=3");
   zdialog_add_widget(zd,"entry","searchcity","hbg1",0,"expand");
   zdialog_add_widget(zd,"label","space","hbg1",0,"space=5");
   zdialog_add_widget(zd,"label","labcountry","hbg1",ZTX("country"),"space=3");
   zdialog_add_widget(zd,"entry","searchcountry","hbg1",0,"expand");

   zdialog_add_widget(zd,"hbox","hbg2","dialog",0);
   zdialog_add_widget(zd,"label","lablat","hbg2","latitude","space=3");
   zdialog_add_widget(zd,"entry","searchlatitude","hbg2",0,"scc=10");
   zdialog_add_widget(zd,"label","space","hbg2",0,"space=5");
   zdialog_add_widget(zd,"label","lablong","hbg2","longitude","space=3");
   zdialog_add_widget(zd,"entry","searchlongitude","hbg2",0,"scc=10");

   zdialog_add_widget(zd,"hbox","hbg3","dialog",0);
   zdialog_add_widget(zd,"button","find","hbg3",Bfind,"space=10");
   zdialog_add_widget(zd,"button","map","hbg3",Bmap,"space=10");
   zdialog_add_widget(zd,"label","labkm","hbg3",ZTX("kilometer range"),"space=10");
   zdialog_add_widget(zd,"entry","searchkilometers","hbg3","100","scc=5");

   zdialog_stuff(zd,"searchcity",searchCity);                              //  stuff prior geotags data
   zdialog_stuff(zd,"searchcountry",searchCountry);
   zdialog_stuff(zd,"searchlatitude",searchLatitude);
   zdialog_stuff(zd,"searchlongitude",searchLongitude);
   zdialog_stuff(zd,"searchkilometers",searchKilometers);
   
   zdialog_run(zd,searchimages_geotags_dialog_event);                      //  run dialog
   zstat = zdialog_wait(zd);                                               //  wait for completion
   zdialog_free(zd);
   
   if (zstat == 1) return 1;                                               //  search criteria were entered
   else return 0;                                                          //  not
}


//  dialog event and completion function

int searchimages_geotags_dialog_event(zdialog *zd, cchar *event)           //  v.12.05
{
   int         fgtags, err, nn;
   char        city[100], country[100];
   char        *location[2], *coord[2];

   if (strEqu(event,"find")) {                                             //  get geotags data from 
      zdialog_fetch(zd,"searchcity",city,99);                              //     (partial) city [country] names
      zdialog_fetch(zd,"searchcountry",country,99);
      location[0] = city;
      location[1] = country;
      nn = geotags_choosecity(location,coord);                             //  find matching city geotags data
      if (nn == 1) {                                                       //    and ask user to choose
         zdialog_stuff(zd,"searchcity",location[0]);
         zdialog_stuff(zd,"searchcountry",location[1]);
         zdialog_stuff(zd,"searchlatitude",coord[0]);
         zdialog_stuff(zd,"searchlongitude",coord[1]);
      }
   }

   if (strEqu(event,"map"))                                                //  get geotags data from world map
   {
      zdialog_show(zd,0);
      get_geotags_worldmap();
      zdialog_show(zd,1);
      
      if (Fgeotags_worldmap_data) {                                        //  use geotags from world map
         zdialog_stuff(zd,"searchcity",geotags_worldmap_city);
         zdialog_stuff(zd,"searchcountry",geotags_worldmap_country);
         zdialog_stuff(zd,"searchlatitude",geotags_worldmap_latitude);
         zdialog_stuff(zd,"searchlongitude",geotags_worldmap_longitude);
         Fgeotags_worldmap_data = 0;
      }
      return 0;
   }

   if (! zd->zstat) return 0;                                              //  wait for completion

   if (zd->zstat != 1) {                                                   //  cancel
      zdialog_destroy(zd);                                                 //  kill dialog
      *searchCity = *searchCountry = 0;                                    //  clear criteria
      *searchLatitude = *searchLongitude = 0;
      return 0;
   }

   zdialog_fetch(zd,"searchcity",searchCity,99);                           //  get search geotags
   zdialog_fetch(zd,"searchcountry",searchCountry,99);
   zdialog_fetch(zd,"searchlatitude",searchLatitude,20);
   zdialog_fetch(zd,"searchlongitude",searchLongitude,20);
   zdialog_fetch(zd,"searchkilometers",searchKilometers);
   
   strToLower(searchCity);
   strToLower(searchCountry);

   err = 0;
   if (*searchLatitude || *searchLongitude) {                              //  sanity check for latitude/longitude
      if (! *searchLatitude || ! *searchLongitude) err++;
      flatitude = atof(searchLatitude);
      flongitude = atof(searchLongitude);
      if (flatitude < -90 || flatitude > 90) err++;
      if (flongitude < -180 || flongitude > 180) err++;
      if (searchKilometers < 1 || searchKilometers > 1000) err++;
      if (strEqu(searchLatitude,"null")) err = 0;                          //  search for null lat/long     v.12.06
   }

   if (err) {
      zmessageACK(mWin,ZTX("geotags search data makes no sense"));
      zd->zstat = 0;
      return 1;
   }

   fgtags = 0;
   if (! blank_null(searchCity)) fgtags = 1;                               //  look for inputs
   if (! blank_null(searchCountry)) fgtags = 1;
   if (! blank_null(searchLatitude)) fgtags = 1;
   if (! blank_null(searchLongitude)) fgtags = 1;

   if (fgtags) zd->zstat = 1;                                              //  search criteria was given
   else zd->zstat = 2;                                                     //  no inputs were made

   zdialog_destroy(zd);                                                    //  kill dialog
   return 1;
}


//  test the image geotags data against the search criteria   

int searchimages_geotags_select(char *gtags)                               //  v.12.05
{
   float       glatitude, glongitude, distance;
   cchar       *pps;

   if (*searchLatitude) {                                                  //  if latitude/longitude present,
      pps = strField(gtags,'^',3);                                         //    ignore city/country data
      if (! pps) return 0;
      if (strEqu(pps,"null"))                                              //  no latitude/longitude data
         if (flatitude == 0 && flongitude == 0) return 1;                  //  this was selected
      glatitude = atof(pps);
      pps = strField(gtags,'^',4);
      if (! pps) return 0;
      if (strEqu(pps,"null")) return 0;
      glongitude = atof(pps);
      distance = (flatitude - glatitude) * (flatitude - glatitude);
      distance += (flongitude - glongitude) * (flongitude - glongitude);
      distance = 111.0 * sqrt(distance);
      if (distance <= searchKilometers) return 1;                          //  within distance limit, select
      return 0;                                                            //  outside, exclude
   }

   if (*searchCity) {                                                      //  check city and country criteria
      pps = strField(gtags,'^',1);
      if (! pps) return 0;
      if (! strcasestr(pps,searchCity)) return 0;
   }
   
   if (*searchCountry) {
      pps = strField(gtags,'^',2);
      if (! pps) return 0;
      if (! strcasestr(pps,searchCountry)) return 0;
   }
   
   return 1;
}


/**************************************************************************/

//  dialog to get metadata search criteria

int searchimages_metadata_dialog()                                         //  v.12.05
{
   int searchimages_metadata_dialog_event(zdialog *zd, cchar *event);

   cchar *metamess = ZTX("These items are always reported: \n"
                         "date, stars, tags, caption, comment");
   zdialog  *zd;
   int      zstat;

/***
         Search Metadata (EXIF/IPTC etc.)

         These items are always reported: 
         date, stars, tags, caption, comment

            Additional Items for Report
            Keyword      Match Criteria
         [__________] [__________________]
         [__________] [__________________]
         [__________] [__________________]
         [__________] [__________________]
         [__________] [__________________]

                       [apply] [cancel]
***/

   zd = zdialog_new("Search Metadata (EXIF/IPTC etc.)",mWin,Bapply,Bcancel,null);
   zdialog_add_widget(zd,"label","labmeta","dialog",metamess,"space=5");
   zdialog_add_widget(zd,"label","labopts","dialog",ZTX("Additional Items for Report"));

   zdialog_add_widget(zd,"hbox","hb1","dialog");
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"space=5|expand");
   
   zdialog_add_widget(zd,"label","lab1","vb1",ZTX("Keyword"));
   zdialog_add_widget(zd,"entry","key0","vb1");
   zdialog_add_widget(zd,"entry","key1","vb1");
   zdialog_add_widget(zd,"entry","key2","vb1");
   zdialog_add_widget(zd,"entry","key3","vb1");
   zdialog_add_widget(zd,"entry","key4","vb1");

   zdialog_add_widget(zd,"label","lab2","vb2",ZTX("Match Criteria"));
   zdialog_add_widget(zd,"entry","match0","vb2",0,"expand");
   zdialog_add_widget(zd,"entry","match1","vb2",0,"expand");
   zdialog_add_widget(zd,"entry","match2","vb2",0,"expand");
   zdialog_add_widget(zd,"entry","match3","vb2",0,"expand");
   zdialog_add_widget(zd,"entry","match4","vb2",0,"expand");
   
   strcpy(searchkeyx,"keyx");
   strcpy(searchkeydatax,"matchx");
   
   if (! searchkeys[0])                                                    //  first call initialization
   {
      for (int ii = 0; ii < 5; ii++) {
         searchkeys[ii] = zmalloc(40,"search_images");
         searchkeydata[ii] = zmalloc(100,"search_images");
         *searchkeys[ii] = *searchkeydata[ii] = 0;
      }
   }   

   for (int ii = 0; ii < Nsearchkeys; ii++)                                //  preload prior data if any
   {
      searchkeyx[3] = '0' + ii;
      if (searchkeys[ii]) zdialog_stuff(zd,searchkeyx,searchkeys[ii]);
      searchkeydatax[5] = '0' + ii;
      if (searchkeydata[ii]) zdialog_stuff(zd,searchkeydatax,searchkeydata[ii]);
   }

   zdialog_resize(zd,400,300);
   zdialog_run(zd,searchimages_metadata_dialog_event);                     //  run dialog
   zstat = zdialog_wait(zd);                                               //  wait for completion
   zdialog_free(zd);
   
   if (zstat == 1) return 1;                                               //  search criteria were entered
   else return 0;                                                          //  not
}


//  dialog event and completion callback function

int searchimages_metadata_dialog_event(zdialog *zd, cchar *event)          //  v.12.05
{
   int      ii, jj;

   if (! zd->zstat) return 1;                                              //  wait for completion

   if (zd->zstat != 1) {
      zdialog_destroy(zd);                                                 //  cancel
      Nsearchkeys = 0;                                                     //  no search keys
      return 1;
   }
   
   Nsearchkeys = 0;

   for (ii = jj = 0; ii < 5; ii++)                                         //  get metadata keys
   {
      searchkeyx[3] = '0' + ii;
      zdialog_fetch(zd,searchkeyx,searchkeys[ii],40);
      strCompress(searchkeys[ii]);                                         //  remove all blanks from key names
      if (*searchkeys[ii] <= ' ') continue;
      memmove(searchkeys[jj],searchkeys[ii],40);                           //  repack blank keys
      searchkeydatax[5] = '0' + ii;
      zdialog_fetch(zd,searchkeydatax,searchkeydata[ii],100);              //  get corresp. match value if any
      strTrim2(searchkeydata[jj],searchkeydata[ii]);                       //  trim leading and trailing blanks
      if (ii > jj) *searchkeys[ii] = *searchkeydata[ii] = 0;
      jj++;
   }

   Nsearchkeys = jj;                                                       //  keys found, no blanks
   
   if (Nsearchkeys) zd->zstat = 1;                                         //  keys were entered
   else zd->zstat = 2;                                                     //  no keys were entered

   zdialog_destroy(zd);                                                    //  kill dialog
   return 1;
}


//  test image metadata against select criteria

int searchimages_metadata_select(char *file)                               //  v.12.05
{
   char           **kvals;
   int            ii, nth;
   cchar          *pp;

   kvals = exif_get(file,(cchar **) searchkeys,Nsearchkeys);               //  get the metadata

   for (ii = 0; ii < Nsearchkeys; ii++) {                                  //  check key values for match
      if (*searchkeydata[ii] > ' ') {                                      //  match value(s) are present
         if (! kvals[ii]) break;                                           //  key not present in metadata
         for (nth = 1; ; nth++) {
            pp = strField(searchkeydata[ii],' ',nth);                      //  get each match value
            if (! pp) break;                                               //  no more, no match found
            if (strcasestr(kvals[ii],pp)) break;                           //  found match
         }
         if (! pp) break;                                                  //  no match found
      }
   }

   if (ii < Nsearchkeys) return 0;                                         //  at least one match fail
   return 1;                                                               //  all matches passed
}


//  Report the selected metadata using a gallery window layout
//  with image thumbnails and selected metadata text.

int searchimages_metadata_report()                                         //  v.12.05
{
   using namespace navi;

   int   metadata_xrec_compare(cchar *rec1, cchar *rec2);
   int   metadata_xrec_search(char *file, xrec_t *xrecs, int nrecs);

   cchar          *keys1[6] = { exif_date_key, iptc_rating_key, exif_city_key, 
                                exif_country_key, iptc_caption_key, exif_comment_key };
   cchar          *keys[20];
   char           *file, **kvals;
   char           text1[2*indexrecl], text2[200];
   int            Nkeys, ii, jj, cc;
   
   if (! mdlist) {
      cc = nfiles * sizeof(char *);                                        //  allocate metadata list
      mdlist = (char **) zmalloc(cc,"search_images");                      //  nfiles = curr. gallery files
      memset(mdlist,0,cc);                                                 //  mdlist = corresp. metadata list
   }                                                                       //  (zfree() in image_navigate)
   
   for (ii = 0; ii < 6; ii++)                                              //  set first 6 key names, fixed
      keys[ii] = keys1[ii];
   
   for (ii = 0; ii < Nsearchkeys; ii++)                                    //  remaining key names from user
      keys[ii+6] = searchkeys[ii];

   Nkeys = 6 + Nsearchkeys;                                                //  total keys to extract

   for (ii = 0; ii < nfiles; ii++)                                         //  loop image gallery files
   {
      file = image_gallery(0,"find",ii);
      if (! file) continue;

      kvals = exif_get(file,(cchar **) keys,Nkeys);                        //  get the metadata
      
      snprintf(text2,200,"%s  %s",kvals[2],kvals[3]);                      //  combine city and country
      if (kvals[2]) zfree(kvals[2]);
      if (kvals[3]) zfree(kvals[3]);
      kvals[2] = strdupz(text2,0,"search_images");
      kvals[3] = 0;

      for (cc = jj = 0; jj < Nkeys; jj++)                                  //  add metadata to report
      {
         if (jj == 3) continue;                                            //  skip country
         snprintf(text2,200,"key: %s  value: %s \n",keys[jj], kvals[jj]);
         strcpy(text1+cc,text2);
         cc += strlen(text2);
      }

      if (mdlist[ii]) zfree(mdlist[ii]);                                   //  attach metadata for gallery paint
      mdlist[ii] = strdupz(text1,0,"search_images");

      for (jj = 0; jj < Nkeys; jj++)                                       //  free memory
         if (kvals[jj]) zfree(kvals[jj]);

      zfree(file);
   }

   if (galleryname) zfree(galleryname);
   galleryname = strdupz("search_images");
   gallerytype = 3;                                                        //  outlaw gallery adds/deletes
   image_gallery(0,"paint1",0,0,mWin);                                     //  show new image gallery window

   return 0;
}


/**************************************************************************/

//  get EXIF/IPTC metadata for given image file and EXIF/IPTC key(s)
//  returns array of pointers to corresponding key values
//  if a key is missing, corresponding pointer is null
//  returned strings belong to caller, are subject for zfree()
//  up to 20 keynames may be requested per call
//  use -fast not -fast2 which can lose geotag data                        v.12.06

char ** exif_get(cchar *file, cchar **keys, int nkeys)                     //  revised       v.12.06
{
   char           *pp, **outputs;
   char           *inputs[30];
   static char    *keyvals[20];
   int            cc, ii, jj;

   if (nkeys < 1 || nkeys > 20) zappcrash("exif_get nkeys: %d",nkeys);

   inputs[0] = (char *) "-m";                                              //  options for exiftool
   inputs[1] = (char *) "-s2";
   inputs[2] = (char *) "-n";
   inputs[3] = (char *) "-fast";
   jj = 4;

   for (ii = 0; ii < nkeys; ii++)                                          //  build exiftool inputs
   {
      cc = strlen(keys[ii]);                                               //  -keyname
      inputs[jj] = zmalloc(cc+2,"exif_get");
      inputs[jj][0] = '-';
      strcpy(inputs[jj]+1,keys[ii]);
      jj++;
   }

   inputs[jj] = strdupz(file,0,"exif_get");                                //  filename last
   jj++;

   inputs[jj] = 0;                                                         //  EOL

   cc = 20 * sizeof(char *);
   memset(keyvals,0,cc);                                                   //  set all outputs to null

   outputs = exiftool_server(inputs);                                      //  get exif outputs
   
   for (ii = 4; ii < jj; ii++)                                             //  free memory
      zfree(inputs[ii]);

   for (ii = 0; outputs[ii]; ii++)                                         //  search outputs
   {
      pp = outputs[ii];                                                    //  keyname: keyvalue

      for (jj = 0; jj < nkeys; jj++)
      {
         uint cc = strlen(keys[jj]);                                       //  look for matching input keyname
         if (strncasecmp(pp,keys[jj],cc) == 0)
            if (strlen(pp) > cc+2)                                         //  if not empty, 
               keyvals[jj] = strdupz(pp+cc+2,0,"exif_get");                //    return keyvalue alone
      }
   }
   
   return keyvals;
}


/**************************************************************************/

//  create or change EXIF/IPTC metadata for given image file and key(s)
//  up to 20 keys may be processed
//  command: 
//    exiftool -overwrite_original -keyname="keyvalue" ... "file"
//
//  NOTE: exiftool replaces \n (newline) in "keyvalue" with "." (period)

int exif_put(cchar *file, cchar **keys, cchar **text, int nkeys)           //  revised          v.12.06
{
   int      ii, jj, cc;
   char     *inputs[30];

   if (nkeys < 1 || nkeys > 20) zappcrash("exif_put nkeys: %d",nkeys);

   inputs[0] = (char *) "-m";                                              //  exiftool options
   inputs[1] = (char *) "-overwrite_original";
   jj = 2;

   for (ii = 0; ii < nkeys; ii++)                                          //  build exiftool inputs
   {
      cc = strlen(keys[ii]) + strlen(text[ii]) + 3;
      inputs[jj] = zmalloc(cc,"exif_put");
      inputs[jj][0] = '-';                                                 //  -keyname=value
      strcpy(inputs[jj]+1,keys[ii]);
      cc = strlen(keys[ii]);
      inputs[jj][cc+1] = '=';
      strcpy(inputs[jj]+cc+2,text[ii]);
      jj++;
      
      if (strcasecmp(keys[ii],"GPSLatitude") == 0) {                       //  take care of latitude N/S 
         if (*text[ii] == '-')
            inputs[jj] = strdupz("-GPSLatitudeRef=S",0,"exif_put");
         else 
            inputs[jj] = strdupz("-GPSLatitudeRef=N",0,"exif_put");
         jj++;
      }

      if (strcasecmp(keys[ii],"GPSLongitude") == 0) {                      //  and longitude E/W
         if (*text[ii] == '-')
            inputs[jj] = strdupz("-GPSLongitudeRef=W",0,"exif_put");
         else 
            inputs[jj] = strdupz("-GPSLongitudeRef=E",0,"exif_put");
         jj++;
      }
   }

   inputs[jj] = strdupz(file,0,"exif_put");                                //  last input is filename
   jj++;
   
   inputs[jj] = 0;                                                         //  EOL

   exiftool_server(inputs);

   for (ii = 2; ii < jj; ii++)                                             //  free memory
       zfree(inputs[ii]);

   return 0;
}


/**************************************************************************/

//  copy EXIF/IPTC data from original image file to new (edited) image file
//  if nkeys > 0, up to 9 keys may be replaced with new values
//  exiftool command revised: parameter sequence and -icc_profile

int exif_copy(cchar *file1, cchar *file2, cchar **keys, cchar **text, int nkeys)
{
   char     *inputs[20];                                                   //  revised       v.12.06
   int      cc, ii, jj;
   
   inputs[0] = (char *) "-m";                                              //  -m
   inputs[1] = (char *) "-tagsfromfile";                                   //  -tagsfromfile
   inputs[2] = strdupz(file1,0,"exif_copy");                               //  file1
   inputs[3] = (char *) "-all";                                            //  -all
   inputs[4] = (char *) "-icc_profile";                                    //  -icc_profile
   
   jj = 5;

   for (int ii = 0; ii < nkeys; ii++)                                      //  -keyname=keyvalue
   {
      cc = strlen(keys[ii]) + strlen(text[ii]) + 3;
      inputs[jj] = zmalloc(cc,"exif_copy");
      strncatv(inputs[jj],cc,"-",keys[ii],"=",text[ii],null);
      jj++;
   }
   
   inputs[jj++] = strdupz(file2,0,"exif_copy");                            //  file2
   inputs[jj++] = (char *) "-overwrite_original";                          //  -overwrite_original
   inputs[jj] = 0;                                                         //  EOL

   exiftool_server(inputs);

   zfree(inputs[2]);                                                       //  free memory
   for (ii = 5; ii < jj-1; ii++)
   zfree(inputs[ii]);

   return 0;
}


/**************************************************************************/

//  convert between EXIF and fotoxx tag date formats
//  EXIF date: yyyy:mm:dd hh:mm:ss[.ss]
//  tag date: yyyymmdd
//  

void exif_tagdate(cchar *exifdate, char *tagdate)
{
   time_t      tnow;
   struct tm   *snow;
   
   if (! exifdate || strlen(exifdate) < 10) {                              //  bad EXIF date, use current date
      tnow = time(0);
      snow = localtime(&tnow);
      snprintf(tagdate,8,"%4d%02d%02d",
               snow->tm_year+1900, snow->tm_mon+1, snow->tm_mday);
      tagdate[8] = 0;
      return;
   }

   strncpy(tagdate,exifdate,4);                                            //  convert 
   strncpy(tagdate+4,exifdate+5,2);
   strncpy(tagdate+6,exifdate+8,2);
   tagdate[8] = 0;
   return;
}

void tag_exifdate(cchar *tagdate, char *exifdate)
{
   int         cc;
   time_t      tnow;
   struct tm   *snow;
   
   if (! tagdate || strlen(tagdate) < 4) {
      tnow = time(0);
      snow = localtime(&tnow);
      snprintf(exifdate,20,"%4d:%02d:%02d %02d:%02d:%02d",
               snow->tm_year+1900, snow->tm_mon+1, snow->tm_mday,
               snow->tm_hour, snow->tm_min, snow->tm_sec);
      exifdate[19] = 0;
      return;
   }

   strncpy(exifdate,tagdate,4);
   strcpy(exifdate+4,":01:01 00:00:00");
   cc = strlen(tagdate);
   if (cc >= 6) strncpy(exifdate+5,tagdate+4,2);
   if (cc >= 8) strncpy(exifdate+8,tagdate+6,2);
   exifdate[19] = 0;
   return;
}


/**************************************************************************

   char ** exiftool_server(char **inputs)
   
   Server wrapper for exiftool for put/get exif/iptc data.
   This saves perl startup overhead for each call (0.1 >> 0.01 secs).

   Input records: -opt1
                  -opt2
                    ...
                  -keyword1=value1
                  -keyword2=value2
                    ... 
                  filename.jpg
                  (null pointer)

   Returned: list of pointers to resulting output records.
   The last output record is a null pointer.
   Returns null if there was a fatal error.
   Prior output records are freed automatically with each 
   new call, so the caller need not worry about this.

   First call: 
      Starts exiftool with pipe input and output files.
      Sends input record to exiftool and returns output records.
   Subsequent calls: 
      The existing exiftool process is re-used so that the 
      substantial startup overhead is avoided.
   To kill the exiftool process: exiftool_server(null).

***************************************************************************/

char ** exiftool_server(char **inputs)                                     //  revised API         v.12.06
{
   int            ii, cc, err;
   static int     fcf = 1;                                                 //  first call flag
   static FILE    *fid1 = 0, *fid2 = 0;                                    //  exiftool input, output files
   char           *pp, command[100];
   static char    *outrecs[100];                                           //  pointers for up to 99 output recs.
   char           outrec[exif_maxcc];                                      //  single output record
   
   if (! inputs)                                                           //  kill exiftool process
   {
      fid1 = fopen(exiftool_input,"a");                                    //  (also orphaned process)
      if (fid1) {
         fprintf(fid1,"-stay_open\nFalse\n");                              //  tell it to exit
         fclose(fid1);
         fflush(fid1);
      }
      remove(exiftool_input);
      if (fid2) pclose(fid2);
      fid2 = 0;
      fcf = 1;                                                             //  start exiftool process if called again
      return 0;
   }

   if (fcf)                                                                //  first call only
   {
      fid1 = fopen(exiftool_input,"w");                                    //  start exiftool input file
      if (! fid1) {
         printf("exiftool_server, fopen error: %s \n",strerror(errno));
         return 0;
      }

      sprintf(command,"exiftool -stay_open True -@ %s",exiftool_input);    //  shell command to start exiftool

      fid2 = popen(command,"r");                                           //  start exiftool and output file
      if (! fid2) {
         printf("exiftool_server, popen error: %s \n",strerror(errno));
         fclose(fid1);
         return 0;
      }

      cc = 100 * sizeof(char *);                                           //  initz. no outputs
      memset(outrecs,0,cc);

      fcf = 0;                                                             //  keep last
   }
   
   for (ii = 0; ii < 99; ii++)                                             //  free memory from prior call
   {
      if (! outrecs[ii]) break;
      zfree(outrecs[ii]);
      outrecs[ii] = 0;
   }
   
   for (ii = 0; inputs[ii]; ii++)
      fprintf(fid1,"%s\n",inputs[ii]);                                     //  write to exiftool, 1 per record

   err = fprintf(fid1,"-execute\n");                                       //  tell exiftool to process
   if (err < 0) {
      printf("exiftool_server, fprintf error: %s \n",strerror(errno));     //  v.12.06
      return 0;
   }

   fflush(fid1);                                                           //  flush buffer

   for (ii = 0; ii < 99; ii++)                                             //  get exiftool outputs
   {
      pp = fgets_trim(outrec,exif_maxcc,fid2,1);
      if (! pp) break;
      if (strncmp(outrec,"{ready}",7) == 0) break;                         //  look for output end
      outrecs[ii] = strdupz(outrec,0,"exiftool_server");                   //  add to returned records
   }

   outrecs[ii] = 0;                                                        //  mark output end
   if (ii == 99) printf("exiftool_server: >99 output recs \n");

   return outrecs;                                                         //  return outputs to caller
}



